//#preprocessor

//#ifndef USE_LCMS_LIBRARY
/*
 * File: CMM.java
 * Version: 1.5
 * Initial Creation: May 20, 2010 9:45:31 PM
 *
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * SEE "Little CMS License.txt"
 */
package com.sun.pdfview.helper.graphics.color;

import java.io.DataInputStream;
import java.io.IOException;

import net.rim.device.api.util.MathUtilities;

import com.sun.pdfview.helper.SeekableStream;

//Why is it called CMM? Should it be named ICC?
/**
 * Used to handle CMM processing.
 * <p>
 * based off Apache source located in their source tar in "working_classlib\modules\awt\src\main\native\lcmm\shared" and Little CMS Engine.
 * @author Vincent Simonetti
 */
public class CMM
{
	//-Constants
	/** Maximum number of channels in ICC profiles*/
	public static final int cmsMAXCHANNELS = 16;
	
	//Shift values
	public static final int FLOAT_SH = 22;
	public static final int COLORSPACE_SH = 16;
	public static final int CHANNELS_SH = 3;
	public static final int BYTES_SH = 0;
	
	// Flags
	/** Inhibit 1-pixel cache*/
	public static final int cmsFLAGS_NOCACHE = 0x0040;
	/** Inhibit optimizations*/
	public static final int cmsFLAGS_NOOPTIMIZE = 0x0100;
	/** Don't transform anyway*/
	public static final int cmsFLAGS_NULLTRANSFORM = 0x0200;
	
	// Proofing flags
	/** Out of Gamut alarm*/
	public static final int cmsFLAGS_GAMUTCHECK = 0x1000;
	/** Do softproofing*/
	public static final int cmsFLAGS_SOFTPROOFING = 0x4000;
	
	// Misc
	public static final int cmsFLAGS_BLACKPOINTCOMPENSATION = 0x2000;
	/** Don't fix scum dot*/
	public static final int cmsFLAGS_NOWHITEONWHITEFIXUP = 0x0004;
	/** Use more memory to give better accuracy*/
	public static final int cmsFLAGS_HIGHRESPRECALC = 0x0400;
	/** Use less memory to minimize resources*/
	public static final int cmsFLAGS_LOWRESPRECALC = 0x0800;
	
	// For devicelink creation
	/** Create 8 bits devicelinks*/
	public static final int cmsFLAGS_8BITS_DEVICELINK = 0x0008;
	/** Guess device class (for transform2devicelink)*/
	public static final int cmsFLAGS_GUESSDEVICECLASS = 0x0020;
	/** Keep profile sequence for devicelink creation*/
	public static final int cmsFLAGS_KEEP_SEQUENCE = 0x0080;
	
	// Specific to a particular optimizations
	/** Force CLUT optimization*/
	public static final int cmsFLAGS_FORCE_CLUT = 0x0002;
	/** create postlinearization tables if possible*/
	public static final int cmsFLAGS_CLUT_POST_LINEARIZATION = 0x0001;
	/** create prelinearization tables if possible*/
	public static final int cmsFLAGS_CLUT_PRE_LINEARIZATION = 0x0010;
	
	// CRD special
	public static final int cmsFLAGS_NODEFAULTRESOURCEDEF = 0x01000000;
	
	// Pixel types
	/** Don't check colorspace*/
	public static final int PT_ANY = 0;
	// 1 & 2 are reserved
	public static final int PT_GRAY = 3;
	public static final int PT_RGB = 4;
	public static final int PT_CMY = 5;
	public static final int PT_CMYK = 6;
	public static final int PT_YCbCr = 7;
	/** Lu'v'*/
	public static final int PT_YUV = 8;
	public static final int PT_XYZ = 9;
	public static final int PT_Lab = 10;
	/** Lu'v'K*/
	public static final int PT_YUVK = 11;
	public static final int PT_HSV = 12;
	public static final int PT_HLS = 13;
	public static final int PT_Yxy = 14;
	
	public static final int PT_MCH1 = 15;
	public static final int PT_MCH2 = 16;
	public static final int PT_MCH3 = 17;
	public static final int PT_MCH4 = 18;
	public static final int PT_MCH5 = 19;
	public static final int PT_MCH6 = 20;
	public static final int PT_MCH7 = 21;
	public static final int PT_MCH8 = 22;
	public static final int PT_MCH9 = 23;
	public static final int PT_MCH10 = 24;
	public static final int PT_MCH11 = 25;
	public static final int PT_MCH12 = 26;
	public static final int PT_MCH13 = 27;
	public static final int PT_MCH14 = 28;
	public static final int PT_MCH15 = 29;
	
	/** Identical to PT_Lab, but using the V2 old encoding*/
	public static final int PT_LabV2 = 30;
	
	/* Mirek: temp (in microreciprocal kelvin)
	 * ut: u coord of intersection w/ blackbody locus
	 * vt: v coord of intersection w/ blackbody locus
	 * tt: slope of ISOTEMPERATURE. line
	 */
	private static final double[][] isotempdata = {
	//  {Mirek, Ut,       Vt,      Tt      } 
	    {0,     0.18006,  0.26352,  -0.24341},
	    {10,    0.18066,  0.26589,  -0.25479},
	    {20,    0.18133,  0.26846,  -0.26876},
	    {30,    0.18208,  0.27119,  -0.28539},
	    {40,    0.18293,  0.27407,  -0.30470},
	    {50,    0.18388,  0.27709,  -0.32675},
	    {60,    0.18494,  0.28021,  -0.35156},
	    {70,    0.18611,  0.28342,  -0.37915},
	    {80,    0.18740,  0.28668,  -0.40955},
	    {90,    0.18880,  0.28997,  -0.44278},
	    {100,   0.19032,  0.29326,  -0.47888},
	    {125,   0.19462,  0.30141,  -0.58204},
	    {150,   0.19962,  0.30921,  -0.70471},
	    {175,   0.20525,  0.31647,  -0.84901},
	    {200,   0.21142,  0.32312,  -1.0182 },
	    {225,   0.21807,  0.32909,  -1.2168 },
	    {250,   0.22511,  0.33439,  -1.4512 },
	    {275,   0.23247,  0.33904,  -1.7298 },
	    {300,   0.24010,  0.34308,  -2.0637 },
	    {325,   0.24702,  0.34655,  -2.4681 },
	    {350,   0.25591,  0.34951,  -2.9641 },
	    {375,   0.26400,  0.35200,  -3.5814 },
	    {400,   0.27218,  0.35407,  -4.3633 },
	    {425,   0.28039,  0.35577,  -5.3762 },
	    {450,   0.28863,  0.35714,  -6.7262 },
	    {475,   0.29685,  0.35823,  -8.5955 },
	    {500,   0.30505,  0.35907,  -11.324 },
	    {525,   0.31320,  0.35968,  -15.628 },
	    {550,   0.32129,  0.36011,  -23.325 },
	    {575,   0.32931,  0.36038,  -40.770 },
	    {600,   0.33724,  0.36051,  -116.45  }
	};
	
	// D50 XYZ normalized to Y=1.0
	private static final double cmsD50X = 0.9642;
	private static final double cmsD50Y = 1;
	private static final double cmsD50Z = 0.8249;
	
	// V4 perceptual black
	private static final double cmsPERCEPTUAL_BLACK_X = 0.00336;
	private static final double cmsPERCEPTUAL_BLACK_Y = 0.0034731;
	private static final double cmsPERCEPTUAL_BLACK_Z = 0.00287;
	
	private static final int MAX_TAG_COUNT = 100;
	//'sf32'
	private static final int cmsSigS15Fixed16ArrayType = 0x73663332;
	//'XYZ '
	private static final int cmsSigXYZType = 0x58595A20;
	
	// Some broken types
	private static final int cmsCorbisBrokenXYZtype = 0x17A505B8;
	
	// Multi process elements types
	// Custom from here, not in the ICC Spec
	// 'l2x '
	static final int cmsSigXYZ2LabElemType = 0x6C327820;
	// 'x2l '
	static final int cmsSigLab2XYZElemType = 0x78326C20;
	// '2 4 '
	static final int cmsSigLabV2toV4 = 0x32203420;
	// '4 2 '
	static final int cmsSigLabV4toV2 = 0x34203220;
    
	// 'idn '
	static final int cmsSigIdentityElemType = 0x69646E20;
	
	//Static fields
	
	private static DescriptorList SupportedTags;
	private static TagHandlerList SupportedTagTypes;
	
	static
	{
		//First setup SupportedTags, backwords so it is easier
		DescriptorList des = new DescriptorList(ICC_Profile.icSigMediaWhitePointTag, new Descriptor(1, new int[]{cmsSigXYZType, cmsCorbisBrokenXYZtype}), null);
		des = new DescriptorList(ICC_Profile.icSigMediaBlackPointTag, new Descriptor(1, new int[]{cmsSigXYZType, cmsCorbisBrokenXYZtype}), des);
		des = new DescriptorList(ICC_Profile.icSigChromaticAdaptationTag, new Descriptor(9, new int[]{cmsSigS15Fixed16ArrayType}), des);
		SupportedTags = des;
		
		//Next do the supported tag types
		TagHandlerList tagType = new TagHandlerList(cmsCorbisBrokenXYZtype, new TagHandler()
		{
			public Object read(CMM cmm, SeekableStream io, int[] nItems, int sizeOfTag)
			{
				nItems[0] = 0;
				
				VEC3 xyz = new VEC3();
				try
				{
					DataInputStream in = new DataInputStream(io);
					xyz.vals[0] = _cms15Fixed16toDouble(reverseEndian(in.readInt()));
					xyz.vals[1] = _cms15Fixed16toDouble(reverseEndian(in.readInt()));
					xyz.vals[2] = _cms15Fixed16toDouble(reverseEndian(in.readInt()));
				}
				catch(IOException ioe)
				{
					return null;
				}
				
				normalizeXYZ(xyz);
				
				nItems[0] = 1;
				return xyz;
			}
		}, null);
		tagType = new TagHandlerList(cmsSigXYZType, tagType.handler, tagType);
		tagType = new TagHandlerList(cmsSigS15Fixed16ArrayType, new TagHandler()
		{
			public Object read(CMM cmm, SeekableStream io, int[] nItems, int sizeOfTag)
			{
				nItems[0] = 0;
				
				int n = sizeOfTag / 4;
				double[] array_double = new double[n];
				
				try
				{
					DataInputStream in = new DataInputStream(io);
					for (int i = 0; i < n; i++)
				    {
						array_double[i] = _cms15Fixed16toDouble(reverseEndian(in.readInt()));     
				    }
				}
				catch(IOException ioe)
				{
					return null;
				}
			    
			    nItems[0] = n;
			    return array_double;
			}
		}, tagType);
		SupportedTagTypes = tagType;
	}
	
	//-Variables
	private int profileSize;
	private Tag[] tags;
	private int[] linkedTags;
	private Object[] loadedTags;
	private TagHandler[] tagHandlers;
	private SeekableStream stream;
	
	private int lastSearchedTagID, lastSearchedTagIndex;
	
	/**
	 * Create a new ICC controller.
	 * @param data The data that defines a ICC.
	 */
	public CMM(byte[] data) throws IOException
	{
		this.stream = new SeekableStream(data);
		
		this.lastSearchedTagID = this.lastSearchedTagIndex = 0;
		
		//Verify data
		byte[] header = new byte[ICC_Profile.headerSize];
		String invalidDataMessage = "Invalid ICC Profile Data";
		
		// Get header from the input stream
		if (this.stream.read(header) != ICC_Profile.headerSize)
		{
			throw new IllegalArgumentException(invalidDataMessage);
		}
		
		// Check the profile data for consistency
		if (CMM.getBigEndianFromByteArray(header, ICC_Profile.icHdrMagic) != ICC_Profile.headerMagicNumber)
		{
			throw new IllegalArgumentException(invalidDataMessage);
		}
		
		// Get profile size from header
		profileSize = CMM.getBigEndianFromByteArray(header, ICC_Profile.icHdrSize);
		byte[] profileData = new byte[profileSize];
		
		// Read the profile itself
		if (this.stream.read(profileData, ICC_Profile.headerSize, profileSize - ICC_Profile.headerSize) != profileSize - ICC_Profile.headerSize)
		{
			throw new IllegalArgumentException(invalidDataMessage);
		}
		
		loadTags();
	}
	
	private void loadTags() throws IOException
	{
		stream.setPosition(ICC_Profile.headerSize);
		DataInputStream din = new DataInputStream(stream);
		
		int tagCount = din.readInt();
		if(tagCount > MAX_TAG_COUNT)
		{
			throw new IllegalArgumentException("Too many tags (" + tagCount + ")");
		}
		int actCount = 0;
		tags = new Tag[tagCount];
		linkedTags = new int[tagCount];
		loadedTags = new Object[tagCount];
		tagHandlers = new TagHandler[tagCount];
		for(int i = 0; i < tagCount; i++)
		{
			tags[i] = new Tag(din.readInt(), din.readInt(), din.readInt());
			//Perform a sanity check
			if(tags[i].offset + tags[i].size > profileSize)
			{
				tags[i] = null;
				continue;
			}
			
			// Search for links
	        for (int j = 0; j < actCount; j++)
	        {
	        	if((tags[j].offset == tags[i].offset) && (tags[j].size == tags[i].size))
	        	{
	        		linkedTags[actCount] = tags[j].ID;
	            }
	        }
	        
	        actCount++;
		}
		if(actCount != tagCount)
		{
			System.err.println("Actual tag count does not match profile tag count. Corrupted ICC profile might be in use.");
		}
	}
	
	private int findTag(int sig)
	{
		//Use cache to speed up searches, for example this would get called by cmmGetProfileElementSize then cmmGetProfileElement, size then data. Why search twice? Cache it.
		if(this.lastSearchedTagID == sig)
		{
			return this.lastSearchedTagIndex;
		}
		int len = this.tags.length;
		for (int i = 0; i < len; i++)
		{
			if (sig == this.tags[i].ID)
			{
				//Save to speed up searches
				this.lastSearchedTagID = sig;
				this.lastSearchedTagIndex = i;
				return i;
			}
		}
		return -1; // Not found
	}
	
	private int _cmsSearchTag(int sig, boolean lFollowLinks)
	{
		int n, LinkedSig;
		
		do
		{
			// Search for given tag in ICC profile directory
			n = findTag(sig);
			if (n < 0)
			{
				return -1;        // Not found
			}
			
			if (!lFollowLinks)
			{
				return n;         // Found, don't follow links
			}
			
			// Is this a linked tag?
			LinkedSig = this.linkedTags[n];
			
			// Yes, follow link
			if (LinkedSig != 0)
			{
				sig = LinkedSig;
			}
		} while(LinkedSig != 0);
		
		return n;
	}
	
	/**
	 * Get a tag size in bytes.
	 */
	public int cmmGetProfileElementSize(int tagSignature)
	{
		int size;
		if(tagSignature == ICC_Profile.icSigHead)
		{
			size = ICC_Profile.headerSize;
		}
		else
		{
			 int idx = findTag(tagSignature);
			 if(idx < 0)
			 {
				 size =  -1;
			 }
			 else
			 {
				 size = this.tags[idx].size;
			 }
		}
		
		if(size < 0)
		{
			throw new RuntimeException("Profile element not found");
		}
		return size;
	}
	
	/**
	 * Get tag data.
	 */
	public void cmmGetProfileElement(int tagSignature, byte[] data) throws IOException
	{
		int len = data.length;
		if(tagSignature == ICC_Profile.icSigHead)
		{
			len = Math.min(len, ICC_Profile.headerSize);
			
			this.stream.setPosition(0);
			this.stream.read(data, 0, len);
		}
		else
		{
			int idx = findTag(tagSignature);
			
			if(idx < 0)
			{
				throw new RuntimeException("Tagged profile element not found");
			}
			
			len = Math.min(len, this.tags[idx].size);
			
			this.stream.setPosition(this.tags[idx].offset);
			this.stream.read(data, 0, len);
		}
	}
	
	private static class Tag
	{
		public int ID;
		public int size;
		public int offset;
		
		public Tag(int ID, int offset, int size)
		{
			this.ID = ID;
			this.size = size;
			this.offset = offset;
		}
	}
	
	private static class Descriptor
	{
		private static final int MAX_TYPES_IN_LCMS_PLUGIN = 20;
		
		public int elemCount;
		public int[] supportedTypes;
		
		public Descriptor(int elemCount, int[] supportedTypes)
		{
			this.elemCount = elemCount;
			this.supportedTypes = supportedTypes;
		}
		
		public boolean isTypeSupported(int sig)
		{
			int len = Math.min(supportedTypes.length, MAX_TYPES_IN_LCMS_PLUGIN);
			for(int i = 0; i < len; i++)
			{
				if(supportedTypes[i] == sig)
				{
					return true;
				}
			}
			return false;
		}
	}
	
	private static class DescriptorList
	{
		public int signature;
		public Descriptor descriptor;
		public DescriptorList next;
		
		public DescriptorList(int sig, Descriptor des, DescriptorList next)
		{
			this.signature = sig;
			this.descriptor = des;
			this.next = next;
		}
	}
	
	private static class TagHandlerList
	{
		public int type;
		public TagHandler handler;
		public TagHandlerList next;
		
		public TagHandlerList(int type, TagHandler handler, TagHandlerList next)
		{
			this.type = type;
			this.handler = handler;
			this.next = next;
		}
	}
	
	// Check existance
	public boolean cmsIsTag(int sig)
	{
		return _cmsSearchTag(sig, false) >= 0;
	}
	
	// Returns TRUE if the profile is implemented as matrix-shaper
	public static boolean cmsIsMatrixShaper(ICC_Profile profile)
	{
		CMM cmm = profile.cmm;
	    switch (cmsGetColorSpace(profile))
	    {
		    case ICC_Profile.icSigGrayData:
		        return cmm.cmsIsTag(ICC_Profile.icSigGrayTRCTag);
		    case ICC_Profile.icSigRgbData:
		        return (cmm.cmsIsTag(ICC_Profile.icSigRedColorantTag) &&
		        		cmm.cmsIsTag(ICC_Profile.icSigGreenColorantTag) &&
		        		cmm.cmsIsTag(ICC_Profile.icSigBlueColorantTag) &&
		        		cmm.cmsIsTag(ICC_Profile.icSigRedTRCTag) &&
		                cmm.cmsIsTag(ICC_Profile.icSigGreenTRCTag) &&
		                cmm.cmsIsTag(ICC_Profile.icSigBlueTRCTag));
		    default:
		        return false;
	    }
	}
	
	// That's the main read function
	public Object cmsReadTag(int sig)
	{
		int n = _cmsSearchTag(sig, true);
		if (n < 0)
		{
			return null; // Not found, return NULL
		}
		
		// If the element is already in memory, return the pointer
		if (this.loadedTags[n] != null)
		{
			return this.loadedTags[n];
		}
		
		// We need to read it. Get the offset and size to the file
	    int Offset = tags[n].offset;
	    int TagSize = tags[n].size;
	    
	    // Seek to its location
	    try
	    {
			this.stream.setPosition(Offset);
		}
	    catch (IOException e)
	    {
	    	return null;
		}
	    
	    // Search for support on this tag
	    Descriptor TagDescriptor = _cmsGetTagDescriptor(sig);
	    if (TagDescriptor == null)
	    {
	    	return null; // Unsupported.
	    }
	    
	    // if supported, get type and check if in list
	    //-This is the function call "_cmsReadTypeBase" inlined
	    int BaseType = 0;
	    try
	    {
	    	DataInputStream dat = new DataInputStream(stream);
		    BaseType = reverseEndian(dat.readInt()); //Sig
		    dat.readInt(); //Reserved
	    }
	    catch(IOException ioe)
	    {
	    }
	    if (BaseType == 0)
	    {
	    	return null;
	    }
	    
	    if (!TagDescriptor.isTypeSupported(BaseType))
	    {
	    	return null;
	    }
	    
	    TagSize -= 8; // Already read by the type base logic
	    
	    // Get type handler
	    TagHandler TypeHandler = _cmsGetTagTypeHandler(BaseType);
	    if (TypeHandler == null)
	    {
	    	return null;
	    }
	    
	    // Read the tag
	    this.tagHandlers[n] = TypeHandler;
	    int[] ElemCount = new int[1];
	    this.loadedTags[n] = TypeHandler.read(this, this.stream, ElemCount, TagSize);
	    
	    // The tag type is supported, but something wrong happend and we cannot read the tag.
	    // let know the user about this (although it is just a warning)
	    if (this.loadedTags[n] == null)
	    {
	    	System.err.println("Corrupted tag '" + CMM._cmsTagSignature2String(sig) + "'");
	    	return null;
	    }
	    
	    // This is a weird error that may be a symptom of something more serious, the number of
	    // stored item is actually less than the number of required elements. 
	    if (ElemCount[0] < TagDescriptor.elemCount)
	    {
	    	StringBuffer buf = new StringBuffer("'");
	    	buf.append(CMM._cmsTagSignature2String(sig));
	    	buf.append("' Inconsistent number of items: expected ");
	    	buf.append(TagDescriptor.elemCount);
	    	buf.append(", got ");
	    	buf.append(ElemCount[0]);
	    	System.err.println(buf.toString());
	    	return null;
	    }
	    
	    // Return the data
	    return this.loadedTags[n];
	}
	
	// Get a media white point fixing some issues found in certain old profiles
	public boolean cmsReadMediaWhitePoint(VEC3 dest)
	{
	    if(dest == null)
	    {
	    	throw new NullPointerException();
	    }
	    
	    byte[] header = new byte[ICC_Profile.headerSize];
	    try
	    {
			cmmGetProfileElement(ICC_Profile.icSigHead, header);
		}
	    catch (IOException e)
	    {
	    	return false;
		}
	    
	    // V2 display profiles should give D50. This was previously after the tag reading but if V2 should get D50 then why do it AFTER reading the tag?
	    if (cmsGetEncodedICCversion(header) < 0x4000000)
	    {
	        if (cmsGetDeviceClass(header) == ICC_Profile.icSigDisplayClass)
	        {
	        	dest.set(cmsD50_XYZ());
		        return true;          
	        }
	    }
	    
	    VEC3 tag = (VEC3)cmsReadTag(ICC_Profile.icSigMediaWhitePointTag);   
	    
	    // If no wp, take D50
	    if (tag == null)
	    {
	    	dest.set(cmsD50_XYZ());
	        return true;
	    }
	    
	    // All seems ok
	    dest.set(tag);
	    return true;
	}
	
	// Chromatic adaptation matrix. Fix some issues as well
	public boolean cmsReadCHAD(MAT3 dest)
	{
	    if(dest == null)
	    {
	    	throw new NullPointerException();
	    }
	    
	    double[] vals = (double[])cmsReadTag(ICC_Profile.icSigChromaticAdaptationTag);
	    if (vals != null)
	    {
	    	dest.set(vals, 0);
	        return true;
	    }
	    
	    // No CHAD available, default it to identity
	    dest.setIdentity();
	    
	    byte[] header = new byte[ICC_Profile.headerSize];
	    try
	    {
			cmmGetProfileElement(ICC_Profile.icSigHead, header);
		}
	    catch (IOException e)
	    {
	    	return false;
		}
	    
	    // V2 display profiles should give D50
	    if (cmsGetEncodedICCversion(header) < 0x4000000)
	    {
	        if (cmsGetDeviceClass(header) == ICC_Profile.icSigDisplayClass)
	        {
	            VEC3 white = (VEC3)cmsReadTag(ICC_Profile.icSigMediaWhitePointTag);   
	            
	            if (white == null)
	            {
	            	//dest.setIdentity(); //dest is set to identity before it gets here
	                return true;
	            }
	            
	            return cmsAdaptationMatrix(dest, null, cmsD50_XYZ(), white);
	        }
	    }
	    
	    return true;
	}
	
	//Static functions
	
	public static int _cmsICCcolorSpace(int OurNotation)
	{
		switch (OurNotation)
		{
			case 1:
			case PT_GRAY:
				return ICC_Profile.icSigGrayData;
			case 2:
			case PT_RGB:
				return ICC_Profile.icSigRgbData;
			case PT_CMY:
				return ICC_Profile.icSigCmyData;
			case PT_CMYK:
				return ICC_Profile.icSigCmykData;
			case PT_YCbCr:
				return ICC_Profile.icSigYCbCrData;
			case PT_YUV:
				return ICC_Profile.icSigLuvData;
			case PT_XYZ:
				return ICC_Profile.icSigXYZData;
			case PT_LabV2:
			case PT_Lab:
				return ICC_Profile.icSigLabData;
			case PT_YUVK:
				return 1282766411; //cmsSigLuvKData
			case PT_HSV:
				return ICC_Profile.icSigHsvData;
			case PT_HLS:
				return ICC_Profile.icSigHlsData;
			case PT_Yxy:
				return ICC_Profile.icSigYxyData;
			case PT_MCH1:
				return 1296255025; //cmsSigMCH1Data
			case PT_MCH2:
				return 1296255026; //cmsSigMCH2Data
			case PT_MCH3:
				return 1296255027; //cmsSigMCH3Data
			case PT_MCH4:
				return 1296255028; //cmsSigMCH4Data
			case PT_MCH5:
				return 1296255029; //cmsSigMCH5Data
			case PT_MCH6:
				return 1296255030; //cmsSigMCH6Data
			case PT_MCH7:
				return 1296255031; //cmsSigMCH7Data
			case PT_MCH8:
				return 1296255032; //cmsSigMCH8Data
			case PT_MCH9:
				return 1296255033; //cmsSigMCH9Data
			case PT_MCH10:
				return 1296255034; //cmsSigMCHAData
			case PT_MCH11:
				return 1296255035; //cmsSigMCHBData
			case PT_MCH12:
				return 1296255036; //cmsSigMCHCData
			case PT_MCH13:
				return 1296255037; //cmsSigMCHDData
			case PT_MCH14:
				return 1296255038; //cmsSigMCHEData
			case PT_MCH15:
				return 1296255039; //cmsSigMCHFData
		}
		return -1;
	}
	
	/**
	 * CMS based function to get the LCMS color type from a colorspace.
	 */
	public static int cmsLCMScolorSpace(int colorSpace)
	{
		switch(colorSpace)
		{
			case ICC_Profile.icSigGrayData:
				return PT_GRAY;
		    case ICC_Profile.icSigRgbData:
		    	return PT_RGB;
		    case ICC_Profile.icSigCmyData:
		    	return PT_CMY;
		    case ICC_Profile.icSigCmykData:
		    	return PT_CMYK;
		    case ICC_Profile.icSigYCbCrData:
		    	return PT_YCbCr;
		    case ICC_Profile.icSigLuvData:
		    	return PT_YUV;
		    case ICC_Profile.icSigXYZData:
		    	return PT_XYZ;
		    case ICC_Profile.icSigLabData:
		    	return PT_Lab;
		    case 1282766411: //cmsSigLuvKData
		    	return PT_YUVK;
		    case ICC_Profile.icSigHsvData:
		    	return PT_HSV;
		    case ICC_Profile.icSigHlsData:
		    	return PT_HLS;
		    case ICC_Profile.icSigYxyData:
		    	return PT_Yxy;
		    case 1296255025: //cmsSigMCH1Data
		    case 826494034: //cmsSig1colorData
		    	return PT_MCH1;
		    case 1296255026: //cmsSigMCH2Data
		    case ICC_Profile.icSigSpace2CLR:
		    	return PT_MCH2;
		    case 1296255027: //cmsSigMCH3Data
		    case ICC_Profile.icSigSpace3CLR:
		    	return PT_MCH3;
		    case 1296255028: //cmsSigMCH4Data
		    case ICC_Profile.icSigSpace4CLR:
		    	return PT_MCH4;
		    case 1296255029: //cmsSigMCH5Data
		    case ICC_Profile.icSigSpace5CLR:
		    	return PT_MCH5;
		    case 1296255030: //cmsSigMCH6Data
		    case ICC_Profile.icSigSpace6CLR:
		    	return PT_MCH6;
		    case 1296255031: //cmsSigMCH7Data
		    case ICC_Profile.icSigSpace7CLR:
		    	return PT_MCH7;
		    case 1296255032: //cmsSigMCH8Data
		    case ICC_Profile.icSigSpace8CLR:
		    	return PT_MCH8;
		    case 1296255033: //cmsSigMCH9Data
		    case ICC_Profile.icSigSpace9CLR:
		    	return PT_MCH9;
		    case 1296255034: //cmsSigMCHAData
		    case ICC_Profile.icSigSpaceACLR:
		    	return PT_MCH10;
		    case 1296255035: //cmsSigMCHBData
		    case ICC_Profile.icSigSpaceBCLR:
		    	return PT_MCH11;
		    case 1296255036: //cmsSigMCHCData
		    case ICC_Profile.icSigSpaceCCLR:
		    	return PT_MCH12;
		    case 1296255037: //cmsSigMCHDData
		    case ICC_Profile.icSigSpaceDCLR:
		    	return PT_MCH13;
		    case 1296255038: //cmsSigMCHEData
		    case ICC_Profile.icSigSpaceECLR:
		    	return PT_MCH14;
		    case 1296255039: //cmsSigMCHFData
		    case ICC_Profile.icSigSpaceFCLR:
		    	return PT_MCH15;
			default:
				return -1;
		}
	}
	
	/**
	 * CMS based function to get the number of channels from a colorspace.
	 */
	public static int cmsChannelsOf(int colorSpace)
	{
		switch(colorSpace)
		{
			case ICC_Profile.icSigGrayData:
				return 1;
		    case ICC_Profile.icSigSpace2CLR:
		    	return 2;
		    default:
		    case ICC_Profile.icSigXYZData:
		    case ICC_Profile.icSigLabData:
		    case ICC_Profile.icSigLuvData:
		    case ICC_Profile.icSigYCbCrData:
		    case ICC_Profile.icSigYxyData:
		    case ICC_Profile.icSigRgbData:
		    case ICC_Profile.icSigHsvData:
		    case ICC_Profile.icSigHlsData:
		    case ICC_Profile.icSigCmyData:
		    case ICC_Profile.icSigSpace3CLR:
		    	return 3;
		    case 1282766411: //cmsSigLuvKData
		    case ICC_Profile.icSigCmykData:
		    case ICC_Profile.icSigSpace4CLR:
		    	return 4;
		    case 1296255029: //cmsSigMCH5Data
		    case ICC_Profile.icSigSpace5CLR:
		    	return 5;
		    case 1296255030: //cmsSigMCH6Data
		    case ICC_Profile.icSigSpace6CLR:
		    	return 6;
		    case 1296255031: //cmsSigMCH7Data
		    case ICC_Profile.icSigSpace7CLR:
		    	return 7;
		    case 1296255032: //cmsSigMCH8Data
		    case ICC_Profile.icSigSpace8CLR:
		    	return 8;
		    case 1296255033: //cmsSigMCH9Data
		    case ICC_Profile.icSigSpace9CLR:
		    	return 9;
		    case 1296255034: //cmsSigMCHAData
		    case ICC_Profile.icSigSpaceACLR:
		    	return 10;
		    case 1296255035: //cmsSigMCHBData
		    case ICC_Profile.icSigSpaceBCLR:
		    	return 11;
		    case 1296255036: //cmsSigMCHCData
		    case ICC_Profile.icSigSpaceCCLR:
		    	return 12;
		    case 1296255037: //cmsSigMCHDData
		    case ICC_Profile.icSigSpaceDCLR:
		    	return 13;
		    case 1296255038: //cmsSigMCHEData
		    case ICC_Profile.icSigSpaceECLR:
		    	return 14;
		    case 1296255039: //cmsSigMCHFData
		    case ICC_Profile.icSigSpaceFCLR:
		    	return 15;
		}
	}
	
	public static int cmsGetColorSpace(byte[] headerData)
	{
		return getIntFromByteArray(headerData, ICC_Profile.icHdrColorSpace);
	}
	
	public static int cmsGetColorSpace(ICC_Profile profile)
	{
		return cmsGetColorSpace(profile.getData(ICC_Profile.icSigHead));
	}
	
	public static int cmsGetDeviceClass(byte[] headerData)
	{
		return getIntFromByteArray(headerData, ICC_Profile.icHdrDeviceClass);
	}
	
	public static int cmsGetDeviceClass(ICC_Profile profile)
	{
		return cmsGetDeviceClass(profile.getData(ICC_Profile.icSigHead));
	}
	
	public static int cmsGetEncodedICCversion(byte[] headerData)
	{
		return getIntFromByteArray(headerData, ICC_Profile.icHdrVersion);
	}
	
	public static int cmsGetEncodedICCversion(ICC_Profile profile)
	{
		return cmsGetEncodedICCversion(profile.getData(ICC_Profile.icSigHead));
	}
	
	public static float cmsGetProfileVersion(byte[] headerData)
	{
		int n = cmsGetEncodedICCversion(headerData) >> 16;
		
	    return baseToBase(n, 16, 10) / 100f;
	}
	
	public static float cmsGetProfileVersion(ICC_Profile profile)
	{
		return cmsGetProfileVersion(profile.getData(ICC_Profile.icSigHead));
	}
	
	// Returns the final chrmatic adaptation from illuminant FromIll to Illuminant ToIll
	// The cone matrix can be specified in ConeMatrix. If NULL, Bradford is assumed
	static boolean cmsAdaptationMatrix(MAT3 r, MAT3 coneMatrix, VEC3 fromIll, VEC3 toIll)
	{
		if (coneMatrix == null)
		{
			coneMatrix = new MAT3( // Bradford matrix
				new VEC3(0.8951, 0.2664, -0.1614),
				new VEC3(-0.7502,  1.7135,  0.0367),
				new VEC3(0.0389, -0.0685,  1.0296)
			);
		}
		
		return computeChromaticAdaptation(r, fromIll, toIll, coneMatrix);	
	}
	
	// Compute chromatic adaptation matrix using Chad as cone matrix
	
	public static boolean computeChromaticAdaptation(MAT3 conversion, VEC3 sourceWhitePoint, VEC3 destWhitePoint, MAT3 chad)
	{
	    MAT3 Tmp = new MAT3(chad);
	    MAT3 Chad_Inv = new MAT3();
	    if(!MAT3.inverse(Tmp, Chad_Inv))
	    {
	    	return false;
	    }
	    
	    VEC3 ConeSourceXYZ = new VEC3(sourceWhitePoint);
	    VEC3 ConeDestXYZ = new VEC3(destWhitePoint);
	    
	    VEC3 ConeSourceRGB = new VEC3();
	    VEC3 ConeDestRGB = new VEC3();
	    MAT3.evaluate(ConeSourceRGB, chad, ConeSourceXYZ);
	    MAT3.evaluate(ConeDestRGB, chad, ConeDestXYZ);
	    
	    // Build matrix
	    MAT3 Cone = new MAT3(
	    		new VEC3(ConeDestRGB.vals[0] / ConeSourceRGB.vals[0], 0, 0),
	    		new VEC3(0, ConeDestRGB.vals[1] / ConeSourceRGB.vals[1], 0),
	    		new VEC3(0, 0, ConeDestRGB.vals[2] / ConeSourceRGB.vals[2])
	    );
	    
	    // Normalize
	    MAT3.multiply(Tmp, Cone, chad);
	    MAT3.multiply(conversion, Chad_Inv, Tmp);
	    
		return true;
	}
	
	// Standard XYZ to Lab. it can handle negative XZY numbers in some cases
	public static void cmsXYZ2Lab(VEC3 whitePoint, VEC3 Lab, VEC3 xyz)
	{
	    if (whitePoint == null)
	    {
	    	whitePoint = cmsD50_XYZ();
	    }
	    
	    double fx = f(xyz.vals[0] / whitePoint.vals[0]);
	    double fy = f(xyz.vals[1] / whitePoint.vals[1]);
	    double fz = f(xyz.vals[2] / whitePoint.vals[2]);
	    
	    Lab.vals[0] = 116 * fy - 16;
	    Lab.vals[1] = 500 * (fx - fy);
	    Lab.vals[2] = 200 * (fy - fz);
	}
	
	// Standard XYZ to Lab. It can return negative XYZ in some cases
	public static void cmsLab2XYZ(VEC3 whitePoint, VEC3 xyz,  VEC3 Lab)
	{
	    if (whitePoint == null)
	    {
	    	whitePoint = cmsD50_XYZ();
	    }
	    
	    double y = (Lab.vals[0] + 16) / 116;
	    double x = y + 0.002 * Lab.vals[1];
	    double z = y - 0.005 * Lab.vals[2];
	    
	    xyz.vals[0] = f_1(x) * whitePoint.vals[0];
	    xyz.vals[1] = f_1(y) * whitePoint.vals[1];
	    xyz.vals[2] = f_1(z) * whitePoint.vals[2];
	}
	
	// Conversions
	public static void cmsXYZ2xyY(VEC3 dest, VEC3 source)
	{
	    double ISum = 1 / (source.vals[0] + source.vals[1] + source.vals[2]);
	    
	    dest.vals[0] = (source.vals[0]) * ISum;
	    dest.vals[1] = (source.vals[1]) * ISum;
	    dest.vals[2] = source.vals[1];
	}
	
	public static void cmsxyY2XYZ(VEC3 dest, VEC3 source)
	{
		dest.vals[0] = (source.vals[0] / source.vals[1]) * source.vals[2];
		dest.vals[1] = source.vals[2];
		dest.vals[2] = ((1 - source.vals[0] - source.vals[1]) / source.vals[1]) * source.vals[2];
	}
	
	private static double f(double t)
	{
	    if (t <= ((24.0/116.0) * (24.0/116.0) * (24.0/116.0)))
	    {
	    	return (841.0/108.0) * t + (16.0/116.0);
	    }
	    else
	    {
	    	return MathUtilities.pow(t, 1.0/3.0); 
	    }
	}
	
	private static double f_1(double t)
	{
	    if (t <= (24.0/116.0))
	    {
	        return (108.0/841.0) * (t - (16.0/116.0));
	    }
	    
	    return t * t * t;
	}
	
	// D50 - Widely used
	public static VEC3 cmsD50_XYZ()
	{
		return new VEC3(cmsD50X, cmsD50Y, cmsD50Z);
	}
	
	// This function shouldn't exist at all -- there is such quantity of broken
	// profiles on black point tag, that we must somehow fix chromaticity to 
	// avoid huge tint when doing Black point compensation. This function does
	// just that. There is a special flag for using black point tag, but turned 
	// off by default because it is bogus on most profiles. The detection algorithm 
	// involves to turn BP to neutral and to use only L component.
	
	public static boolean cmsDetectBlackPoint(VEC3 blackPoint, ICC_Profile profile, int Intent, int dwFlags)
	{
	    // Zero for black point
	    if (cmsGetDeviceClass(profile) == ICC_Profile.icSigLinkClass)
	    {
	    	blackPoint.set(0, 0, 0);
	        return false;
	    }
	    
	    // v4 + perceptual & saturation intents does have its own black point, and it is 
	    // well specified enough to use it. Black point tag is deprecated in V4.
	    
	    byte[] header = profile.getData(ICC_Profile.icSigHead);
	    if ((cmsGetEncodedICCversion(header) >= 0x4000000) && (Intent == ICC_Profile.icPerceptual || Intent == ICC_Profile.icSaturation))
	    {
	    	// Matrix shaper share MRC & perceptual intents
	    	if (cmsIsMatrixShaper(profile))
	    	{
	    		return blackPointAsDarkerColorant(profile, ICC_Profile.icRelativeColorimetric, blackPoint, 0);
	    	}
	    	
	    	// Get Perceptual black out of v4 profiles. That is fixed for perceptual & saturation intents
	    	blackPoint.set(cmsPERCEPTUAL_BLACK_X, cmsPERCEPTUAL_BLACK_Y, cmsPERCEPTUAL_BLACK_Z);
	    	
	    	return true;
	    }
//#endif
//#ifdef CMS_USE_PROFILE_BLACK_POINT_TAG
	    
	    // v2, v4 rel/abs colorimetric
	    CMM cmm = profile.cmm;
	    if (cmm.cmsIsTag(ICC_Profile.icSigMediaBlackPointTag) && Intent == ICC_Profile.icRelativeColorimetric)
	    {
	    	// If black point is specified, then use it,
	    	
	    	VEC3 BlackPtr = (VEC3)cmm.cmsReadTag(ICC_Profile.icSigMediaBlackPointTag);
	    	if (BlackPtr != null)
	    	{
	    		VEC3 BlackXYZ = new VEC3(BlackPtr);
	    		VEC3 MediaWhite = new VEC3();
	    		cmm.cmsReadMediaWhitePoint(MediaWhite);
	    		
	    		// Black point is absolute XYZ, so adapt to D50 to get PCS value
	    		VEC3 UntrustedBlackPoint = new VEC3();
	    		cmsAdaptToIlluminant(UntrustedBlackPoint, MediaWhite, cmsD50_XYZ(), BlackXYZ);
	    		
	    		VEC3 Lab = new VEC3();
	    		// Force a=b=0 to get rid of any chroma
	    		cmsXYZ2Lab(null, Lab, UntrustedBlackPoint);
	    		Lab.vals[1] = Lab.vals[2] = 0;
	    		if (Lab.vals[0] > 50)
	    		{
	    			Lab.vals[0] = 50; // Clip to L* <= 50
	    		}
	    		VEC3 TrustedBlackPoint = new VEC3();
	    		cmsLab2XYZ(null, TrustedBlackPoint, Lab);
	    		
	    		if (blackPoint != null)
	    		{
	    			blackPoint.set(TrustedBlackPoint);
	    		}
	    		
	    		return true;
	    	}
	    }
//#endif
//#ifndef USE_LCMS_LIBRARY
	    // That is about v2 profiles. 
	    
	    // If output profile, discount ink-limiting and that's all
	    if (Intent == ICC_Profile.icRelativeColorimetric && (cmsGetDeviceClass(header) == ICC_Profile.icSigOutputClass) && 
	    		(cmsGetColorSpace(header)  == ICC_Profile.icSigCmykData))
	    {
	    	return blackPointUsingPerceptualBlack(blackPoint, profile);
	    }
	    
	    // Nope, compute BP using current intent.
	    return blackPointAsDarkerColorant(profile, Intent, blackPoint, dwFlags);
	}
	
	// Obtains WhitePoint from Temperature
	public static boolean cmsWhitePointFromTemp(VEC3 whitePoint, double tempK)
	{
		if(whitePoint == null)
		{
			throw new NullPointerException();
		}
		
		double T = tempK;
		double T2 = T * T; // Square
		double T3 = T2 * T; // Cube
		
		double x;
		// For correlated color temperature (T) between 4000K and 7000K:
		
		if (T >= 4000 && T <= 7000)
		{
			x = -4.6070 * (1E9 / T3) + 2.9678 * (1E6 / T2) + 0.09911 * (1E3 / T) + 0.244063;
		}
		else
		{
			// or for correlated color temperature (T) between 7000K and 25000K:
			
			if (T > 7000 && T <= 25000)
			{
				x = -2.0064 * (1E9 / T3) + 1.9018 * (1E6 / T2) + 0.24748 * (1E3 / T) + 0.237040;
			}
			else
			{
				throw new IllegalArgumentException("cmsWhitePointFromTemp: invalid temp");
			}
		}
		
		// Obtain y(x)
		
		double y = -3 * (x * x) + 2.870 * x - 0.275;
		
		// wave factors (not used, but here for futures extensions)
		
		// double M1 = (-1.3515 - 1.7703 * x + 5.9114 * y) / (0.0241 + 0.2562 * x - 0.7341 * y);
		// double M2 = (0.0300 - 31.4424 * x + 30.0717 * y) / (0.0241 + 0.2562 * x - 0.7341 * y);
		
		whitePoint.vals[0] = x;
		whitePoint.vals[1] = y;
		whitePoint.vals[2] = 1;
		
		return true;
	}
	
	// Use darker colorants to obtain black point. This works in the relative colorimetric intent and
	// assumes more ink results in darker colors. No ink limit is assumed.
	public static boolean blackPointAsDarkerColorant(ICC_Profile hInput, int Intent, VEC3 BlackPoint, int dwFlags)
	{
		//XXX BlackPointAsDarkerColorant:cmssamp.c:#62
		throw new UnsupportedOperationException("CMM.blackPointAsDarkerColorant not implemented");
	}
	
	// Get a black point of output CMYK profile, discounting any ink-limiting embedded 
	// in the profile. For doing that, we use perceptual intent in input direction:
	// Lab (0, 0, 0) -> [Perceptual] Profile -> CMYK -> [Rel. colorimetric] Profile -> Lab
	public static boolean blackPointUsingPerceptualBlack(VEC3 BlackPoint, ICC_Profile hProfile)
	{
		//XXX BlackPointUsingPerceptualBlack:cmssamp.c:#146
		throw new UnsupportedOperationException("CMM.blackPointUsingPerceptualBlack not implemented");
	}
	
	// Robertson's methodea
	public static boolean cmsTempFromWhitePoint(double[] tempK, VEC3 whitePoint)
	{
		if(tempK == null || tempK.length == 0)
		{
			throw new IllegalArgumentException("tempK");
		}
	    double mi;
		double di = mi = 0;
		double xs = whitePoint.vals[0];
		double ys = whitePoint.vals[1];
		
		// convert (x,y) to CIE 1960 (u,WhitePoint) 
		
		double us = (2 * xs) / (-xs + 6 * ys + 1.5);
		double vs = (3 * ys) / (-xs + 6 * ys + 1.5);
		
		double dj;
		
		int NISO = isotempdata.length;
		for (int j = 0; j < NISO; j++)
		{
			double uj = isotempdata[j][1]; //ut
			double vj = isotempdata[j][2]; //vt
			double tj = isotempdata[j][3]; //tt
			double mj = isotempdata[j][0]; //Mirek

			dj = ((vs - vj) - tj * (us - uj)) / Math.sqrt(1 + tj * tj);
			
			if ((j != 0) && (di/dj < 0.0))
			{
				// Found a match
				tempK[0] = 1000000 / (mi + (di / (di - dj)) * (mj - mi));
				return true;
			}
			
			di = dj;
			mi = mj;
		}
		
		// Not found
		return false;
	}
	
	// Adapts a color to a given illuminant. Original color is expected to have
	// a SourceWhitePt white point. 
	public static boolean cmsAdaptToIlluminant(VEC3 Result, VEC3 SourceWhitePt, VEC3 Illuminant, VEC3 Value)
	{
		if(Result == null || SourceWhitePt == null || Illuminant == null || Value == null)
		{
			throw new NullPointerException();
		}
		
		MAT3 Bradford = new MAT3();
		if (!cmsAdaptationMatrix(Bradford, null, SourceWhitePt, Illuminant))
		{
			return false;
		}
		
		VEC3 In = new VEC3(Value);
		VEC3 Out = new VEC3();
		MAT3.evaluate(Out, Bradford, In);
		
		Result.set(Out);
		
		return true;
	}
	
	private static Descriptor _cmsGetTagDescriptor(int sig)
	{
		for (DescriptorList pt = SupportedTags; pt != null; pt = pt.next)
		{
			if (sig == pt.signature)
			{
				return pt.descriptor;
			}
		}
		return null;
	}
	
	// Wrapper for tag types
	private static TagHandler _cmsGetTagTypeHandler(int sig)
	{
	    return getHandler(sig, SupportedTagTypes);
	}
	
	// Return handler for a given type or NULL if not found. Shared between normal types and MPE
	private static TagHandler getHandler(int sig, TagHandlerList LinkedList)
	{
	    for (TagHandlerList pt = LinkedList; pt != null; pt = pt.next)
	    {
	    	if(pt.type == sig)
	    	{
	    		return pt.handler;
	    	}
	    }
	    
	    return null;
	}
	
	// from Fixed point 15.16 to double
	private static double _cms15Fixed16toDouble(int fix32)
	{
	    int sign  = (fix32 < 0 ? -1 : 1);
	    fix32 = Math.abs(fix32);
	    
	    int Whole = (fix32 >> 16) & 0xFFFF;
	    int FracPart  = (fix32 & 0xFFFF);
	    
	    double mid = FracPart / 65536.0;
	    double floater = Whole + mid;
	    
	    return sign * floater;
	}
	
	// Jun-21-2000: Some profiles (those that comes with W2K) comes
	// with the media white (media black?) x 100. Add a sanity check

	private static void normalizeXYZ(VEC3 dest)
	{
	    while (dest.vals[0] > 2 && dest.vals[1] > 2 && dest.vals[2] > 2)
	    {
	    	dest.vals[0] /= 10;
	    	dest.vals[1] /= 10;
	    	dest.vals[2] /= 10;
	    }
	}
	
	//Includes utility methods for reading ICC profile data.
	
	/**
	 * Utility method.
	 * Gets integer value from the byte array
	 * @param byteArray - byte array
	 * @param idx - byte offset
	 * @return integer value
	 */
	public static int getIntFromByteArray(byte[] byteArray, int idx)
	{
		return ( (byteArray[idx] & 0xFF) << 24)		|
				((byteArray[idx+1] & 0xFF) << 16)	|
				((byteArray[idx+2] & 0xFF) << 8)	|
				( byteArray[idx+3] & 0xFF);
	}
	
	/**
	 * Utility method.
	 * Gets big endian integer value from the byte array
	 * @param byteArray - byte array
	 * @param idx - byte offset
	 * @return integer value
	 */
	public static int getBigEndianFromByteArray(byte[] byteArray, int idx)
	{
		return ( (byteArray[idx] & 0xFF) << 24)		|
				((byteArray[idx+1] & 0xFF) << 16)	|
				((byteArray[idx+2] & 0xFF) << 8)	|
				( byteArray[idx+3] & 0xFF);
	}
	
	/**
     * Utility method.
     * Gets short value from the byte array
     * @param byteArray - byte array
     * @param idx - byte offset
     * @return short value
     */
    public static short getShortFromByteArray(byte[] byteArray, int idx)
    {
        return (short)((byteArray[idx] & 0xFF) | ((byteArray[idx+1] & 0xFF) << 8));
    }
    
    public static int reverseEndian(int value)
    {
    	int nV = (value & 0xFF) << 24;
    	nV |= (value & 0xFF00) << 8;
    	nV |= (value & 0xFF0000) >> 8;
    	return nV | (value & 0xFF000000) >>> 24;
    }
    
    public static short reverseEndian(short value)
    {
    	int nV = (value & 0xFF) << 8;
    	return (short)(nV | ((value & 0xFF00) >> 8));
    }
    
    private static String _cmsTagSignature2String(int sig)
    {
    	StringBuffer buf = new StringBuffer(4);
    	
    	while(sig != 0)
    	{
    		buf.append((char)(sig % 16));
    		sig /= 16;
    	}
    	buf.reverse();
    	
    	return buf.toString();
    }
    
    // Get an hexadecimal number with same digits as v
    private static int baseToBase(int in, int baseIn, int baseOut)
    {
        byte[] buff = new byte[100];
        int out, len, i;
        
        for(len = 0; in > 0 && len < 100; len++)
        {
        	buff[len] = (byte)(in % baseIn);
            in /= baseIn;
        }
        
        for(i = len - 1, out = 0; i >= 0; --i)
        {
            out = out * baseOut + buff[i];
        }
        
        return out;
    }

}
//#endif
