//#preprocessor

//#ifndef USE_LCMS_LIBRARY

/*
 * File: ICC_Pipeline.java
 * Version: 2.0
 * Initial Creation: Jul 3, 2010 5:28:54 PM
 *
 * SEE "Little CMS License.txt"
 */
package com.sun.pdfview.helper.graphics.color;

/**
 * LCMS based pipeline structure
 */
public class ICC_Pipeline
{
	private static final int READ_DEVICE_LINK = 1;
	private static final int READ_INPUT = 2;
	private static final int READ_OUTPUT = 3;
	
	private static OptimizationCollection optimizationCollection = null;
	
	//Points to elements chain
	private ICC_Stage elements;
	public int inputChannels, outputChannels;
	private Object data;
	Evaluate16 eval16Fn;
	EvaluateFloat evalFloatFn;
	
	private ICC_Pipeline(int inputChannels, int outputChannels)
	{
		if(inputChannels >= CMM.cmsMAXCHANNELS || outputChannels >= CMM.cmsMAXCHANNELS)
		{
			throw new IllegalArgumentException("Too many channels, " + CMM.cmsMAXCHANNELS + " is max.");
		}
		
		this.inputChannels = inputChannels;
		this.outputChannels = outputChannels;
		
		this.data = this;
		
		evaluationSetup();
		
		bless();
	}
	
	private ICC_Pipeline(ICC_Profile profile, int intent, int readType)
	{
		//READ_DEVICE_LINK: _cmsReadDevicelinkLUT:cmsio1.c:451
		//READ_INPUT: _cmsReadInputLUT:cmsio1.c:244
		//READ_OUTPUT: _cmsReadOutputLUT:cmsio1.c:392
		
		//TODO
	}
	
	private void evaluationSetup()
	{
		this.eval16Fn = new Evaluate16()
		{
			//Default to evaluate the LUT on 16 bit-basis. Precision is retained.
			public void evaluate(short[] in, short[] out)
			{
				float[][] storage = new float[2][Math.max(ICC_Pipeline.this.inputChannels, ICC_Pipeline.this.outputChannels)];
			    int phase = 0, nextPhase;
			    
			    from16ToFloat(in, storage[phase], ICC_Pipeline.this.inputChannels);

			    for (ICC_Stage mpe = ICC_Pipeline.this.elements; mpe != null; mpe = mpe.next())
			    {
			             nextPhase = phase ^ 1;
			             mpe.eval(storage[phase], storage[nextPhase]);
			             phase = nextPhase;
			    }
			    
			    fromFloatTo16(storage[phase], out, ICC_Pipeline.this.outputChannels);
			}
		};
		this.evalFloatFn = new EvaluateFloat()
		{
			//Does evaluate the LUT on cmsFloat32Number-basis.
			public void evaluate(float[] in, float[] out)
			{
				float[][] storage = new float[2][Math.max(ICC_Pipeline.this.inputChannels, ICC_Pipeline.this.outputChannels)];
			    int phase = 0, nextPhase;
			    
			    System.arraycopy(in, 0, storage[phase], 0, ICC_Pipeline.this.inputChannels);
			    
			    for (ICC_Stage mpe = ICC_Pipeline.this.elements; mpe != null; mpe = mpe.next())
			    {
			              nextPhase = phase ^ 1;
			              mpe.eval(storage[phase], storage[nextPhase]);
			              phase = nextPhase;
			    }
			    
			    System.arraycopy(storage[phase], 0, out, 0, ICC_Pipeline.this.outputChannels);
			}
		};
	}
	
	private void bless()
	{
		if(this.elements != null)
		{
			ICC_Stage first = cmsPipelineGetPtrToFirstStage();
			ICC_Stage last = cmsPipelineGetPtrToLastStage();
			
			if (first != null)
			{
				this.inputChannels = first.inputChannels;
			}
	        if (last != null)
	        {
	        	this.outputChannels = last.outputChannels;
	        }
		}
	}
	
	private ICC_Stage cmsPipelineGetPtrToFirstStage()
	{
		return this.elements;
	}
	
	private ICC_Stage cmsPipelineGetPtrToLastStage()
	{
		ICC_Stage anterior = null;
		
	    for (ICC_Stage mpe = this.elements; mpe != null; mpe = mpe.next())
	    {
	    	anterior = mpe;
	    }
	    
	    return anterior;
	}
	
	// Unlink an element and return the pointer to it
	private void cmsPipelineUnlinkStage(boolean insertAtEnd, ICC_Stage[] mpe)
	{
		ICC_Stage Anterior, pt, Last;
		ICC_Stage Unlinked = null;
		
	    // If empty LUT, there is nothing to remove
		if (this.elements == null)
		{
			if(mpe != null)
			{
				mpe[0] = null;
			}
			return;
		}
		
	    // On depending on the strategy...
		if (insertAtEnd)
		{
			Anterior = Last = null;
			for (pt = this.elements; pt != null; pt = pt.next())
			{
				Anterior = Last;
				Last = pt;
			}
			
			Unlinked = Last; // Next already points to NULL
			
			// Truncate the chain
            if (Anterior != null)
            {
            	Anterior.setNext(null);
            }
            else
            {
            	this.elements = null;
            }
		}
		else
		{
			ICC_Stage elem = this.elements;
			
			this.elements = elem.next();
			elem.setNext(null);
			Unlinked = elem;
		}
		
		if(mpe != null)
		{
			mpe[0] = Unlinked;
		}
		
		bless();
	}
	
	public boolean concatenate(ICC_Pipeline pipeline)
	{
		// If both LUTS does not have elements, we need to inherit
	    // the number of channels
		if(this.elements == null && pipeline.elements == null)
		{
			this.inputChannels = pipeline.inputChannels;
			this.outputChannels = pipeline.outputChannels;
		}
		
		// Cat second
		for(ICC_Stage mpe = pipeline.elements; mpe != null; mpe = mpe.next())
		{
			// We have to dup each element
			ICC_Stage newMPE = mpe.clone();
			
			if(newMPE == null)
			{
				return false;
			}
			
			insertStage(newMPE, true);
		}
		
		bless();
		return true;
	}
	
	public void insertStage(ICC_Stage stage, boolean insertAtEnd)
	{
		if(insertAtEnd)
		{
			if(this.elements == null)
			{
				this.elements = stage;
			}
			else
			{
				ICC_Stage anterior = null;
				for (ICC_Stage pt = this.elements; pt != null; pt = pt.next())
				{
					anterior = pt;
				}
				anterior.setNext(stage);
				stage.setNext(null);
			}
		}
		else
		{
			stage.setNext(this.elements);
			this.elements = stage;
		}
	    
	    bless();
	}
	
	public void removeStage(ICC_Stage stage)
	{
		ICC_Stage prev = getPreviousStage(stage);
		if(prev == null)
		{
			return; //No previous element found or stage not part of pipeline
		}
		prev.setNext(stage.next());
	}
	
	public int stageCount()
	{
		int n = 0;
		for (ICC_Stage mpe = this.elements; mpe != null; mpe = mpe.next(), n++);
		return n;
	}
	
	private ICC_Stage getPreviousStage(ICC_Stage stage)
	{
		ICC_Stage prev;
		for(prev = this.elements; prev != null; prev = prev.next())
		{
			if(prev.next() == stage) //Found the previous stage
			{
				return prev;
			}
		}
		return null;
	}
	
	// Remove all identities in chain. Note that pt actually is a double pointer to the element that holds the pointer. 
	private boolean _remove1Op(int unaryOp)
	{   
		ICC_Stage pt = this.elements;
		boolean anyOpt = false;
		
		while(pt != null)
		{
			if(pt.stageImplements == unaryOp)
			{
				removeStage(pt);
				anyOpt = true;
			}
			else
			{
				pt = pt.next();
			}
		}
		
		return anyOpt;
	}
	
	// Same, but only if two adjacent elements are found
	private boolean _remove2Op(int op1, int op2)
	{
		boolean anyOpt = false;
		
		ICC_Stage pt1 = this.elements;
		ICC_Stage pt2;
		if(pt1 == null)
		{
			return anyOpt;
		}
		
		while(pt1 != null)
		{
			pt2 = pt1.next();
			if(pt2 == null)
			{
				return anyOpt;
			}
			
			if(pt1.stageImplements == op1 && pt2.stageImplements == op2)
			{
				removeStage(pt2);
				removeStage(pt1);
				anyOpt = true;
			}
			else
			{
				pt1 = pt1.next();
			}
		}
		
		return anyOpt;
	}
	
	// Preoptimize just gets rif of no-ops coming paired. Conversion from v2 to v4 followed 
	// by a v4 to v2 and vice-versa. The elements are then discarded.
	private boolean preOptimize()
	{
		boolean anyOpt = false;
		boolean opt;
		
		do
		{
			opt = false;
			
			// Remove all identities
	        opt |= _remove1Op(CMM.cmsSigIdentityElemType);
	        
	        // Remove XYZ2Lab followed by Lab2XYZ
	        opt |= _remove2Op(CMM.cmsSigXYZ2LabElemType, CMM.cmsSigLab2XYZElemType);
	        
	        // Remove Lab2XYZ followed by XYZ2Lab
	        opt |= _remove2Op(CMM.cmsSigLab2XYZElemType, CMM.cmsSigXYZ2LabElemType);
	        
	        // Remove V4 to V2 followed by V2 to V4
	        opt |= _remove2Op(CMM.cmsSigLabV4toV2, CMM.cmsSigLabV2toV4);
	        
	        // Remove V2 to V4 followed by V4 to V2
	        opt |= _remove2Op(CMM.cmsSigLabV2toV4, CMM.cmsSigLabV4toV2);
	        
	        if(opt)
	        {
	        	anyOpt = true;
	        }
		} while(opt);
		
		return anyOpt;
	}
	
	// The entry point for LUT optimization
	public boolean _cmsOptimizePipeline(int intent, int inputFormat, int outputFormat, int[] flags)
	{
		boolean anySuccess = false;
		
		// A CLUT is being asked, so force this specific optimization
	    if ((flags[0] & CMM.cmsFLAGS_FORCE_CLUT) != 0)
	    {
	        preOptimize();
	        return optimizeByResampling(new ICC_Pipeline[]{this}, intent, inputFormat, outputFormat, flags);
	    }
	    
	    Evaluate16 fastIdentity16 = new Evaluate16()
	    {
			public void evaluate(short[] in, short[] out)
			{
				for (int i = 0; i < ICC_Pipeline.this.inputChannels; i++)
				{
			         out[i] = in[i];
			    }
			}
		};
	    
	    // Anything to optimize?
	    if(this.elements == null)
	    {
	        _cmsPipelineSetOptimizationParameters(this, fastIdentity16, this);
	        return true;
	    }
	    
	    // Try to get rid of identities and trivial conversions.
	    anySuccess = preOptimize();
	    
	    // After removal do we end with an identity?
	    if(this.elements == null)
	    {
	    	_cmsPipelineSetOptimizationParameters(this, fastIdentity16, this);
	        return true;
	    }
	    
	    // Do not optimize, keep all precision
	    if((flags[0] & CMM.cmsFLAGS_NOOPTIMIZE) != 0)
	    {
	    	return false;
	    }
	    
	    // Try built-in optimizations and plug-in
	    System.out.println("Warning: ICC_Pipeline._cmsOptimizePipeline gets to actual optimization function, no functions are implemented. Need to implement.");
	    for (OptimizationCollection opts = optimizationCollection; opts != null; opts = opts.next)
	    {
	    	// If one schema succeeded, we are done
	    	if(opts.opt.optimize(this, intent, inputFormat, outputFormat, flags[0]))
	    	{
	    		return true; // Optimized!
	    	}
	    }
		
		// Only simple optimizations succeeded
	    return anySuccess;
	}
	
	//Static functions
	
	// -----------------------------------------------------------------------------------------------------------------------------------------------
	// This function creates simple LUT from complex ones. The generated LUT has an optional set of 
	// prelinearization curves, a CLUT of nGridPoints and optional postlinearization tables. 
	// These curves have to exist in the original LUT in order to be used in the simplified output. 
	// Caller may also use the flags to allow this feature.
	// LUTS with all curves will be simplified to a single curve. Parametric curves are lost.
	// This function should be used on 16-bits LUTS only, as floating point losses precision when simplified
	// -----------------------------------------------------------------------------------------------------------------------------------------------

	private static boolean optimizeByResampling(ICC_Pipeline[] lut, int intent, int inputFormat, int outputFormat, int[] flags)
	{
		// This is a loosy optimization! does not apply in floating-point cases
	    if (ICC_Converter._cmsFormatterIsFloat(inputFormat) || ICC_Converter._cmsFormatterIsFloat(outputFormat))
	    {
	    	return false;
	    }
	    
	    int colorSpace = CMM._cmsICCcolorSpace((inputFormat >> CMM.COLORSPACE_SH) & 0x1F);
	    int outputColorSpace = CMM._cmsICCcolorSpace((outputFormat >> CMM.COLORSPACE_SH) & 0x1F);
	    int gridPoints = _cmsReasonableGridpointsByColorspace(colorSpace, flags[0]);
	    
	    // For empty LUTs, 2 points are enough
	    if (lut[0].stageCount() == 0)
	    {
	    	gridPoints = 2;
	    }
	    
	    ICC_Pipeline src = lut[0];
	    
	    // Allocate an empty LUT    
	    ICC_Pipeline dest = new ICC_Pipeline(src.inputChannels, src.outputChannels);
	    
	    ICC_Stage[] keepPreLin = new ICC_Stage[1];
	    
	    // Prelinearization tables are kept unless indicated by flags
	    if ((flags[0] & CMM.cmsFLAGS_CLUT_PRE_LINEARIZATION) != 0)
	    {
	        // Get a pointer to the prelinearization element
	    	ICC_Stage preLin = src.cmsPipelineGetPtrToFirstStage();
	    	
	        // Check if suitable
	    	if (preLin != null && preLin.type == ICC_Stage.cmsSigCurveSetElemType)
	    	{
	            // Maybe this is a linear tram, so we can avoid the whole stuff
	            if (!allCurvesAreLinear(preLin))
	            {
	                // All seems ok, proceed.
	            	ICC_Stage newPreLin = preLin.clone();
	            	dest.insertStage(newPreLin, false);
	            	
	                // Remove prelinearization. Since we have duplicated the curve
	                // in destination LUT, the sampling shoud be applied after this stage.
	            	src.cmsPipelineUnlinkStage(false, keepPreLin);
	            }
	        }
	    }
	    
	    ICC_Stage CLUT = ICC_Stage.CLUT_Stage.createCLUT_Stage(gridPoints, src.inputChannels, src.outputChannels, null, true);
	    
		//TODO: OptimizeByResampling:cmsopt.c:#582
		return true;
	}
	
	// Try to see if the curves of a given MPE are linear
	private static boolean allCurvesAreLinear(ICC_Stage mpe)
	{
		if(mpe instanceof ICC_Stage.ToneCurves_Stage)
		{
			ICC_Stage.ToneCurves_Stage cMpe = (ICC_Stage.ToneCurves_Stage)mpe;
			
			ICC_Stage.ToneCurves_Stage.ToneCurve[] Curves = cMpe._cmsStageGetPtrToCurveSet();
			
			int n = mpe.outputChannels;
			for(int i = 0; i < n; i++)
			{
				if(!Curves[i].isToneCurveLinear())
				{
					return false;
				}
			}
			
			return true;
		}
		return false;
	}
	
	// This function returns a number of gridpoints to be used as LUT table. It assumes same number
	// of gridpoints in all dimensions. Flags may override the choice.
	private static int _cmsReasonableGridpointsByColorspace(int colorspace, int flags)
	{
	    // Already specified?
	    if ((flags & 0x00FF0000) != 0)
	    {
	    	// Yes, grab'em
	    	return (flags >> 16) & 0xFF;
	    }
	    
	    int nChannels = CMM.cmsChannelsOf(colorspace);
	    
	    // HighResPrecalc is maximum resolution
	    if ((flags & CMM.cmsFLAGS_HIGHRESPRECALC) != 0)
	    {
	        if (nChannels > 4)
	        {
	        	return 7; // 7 for Hifi
	        }
	        
	        if (nChannels == 4)
	        {
	        	return 23; // 23 for CMYK
	        }
	        
	        return 49; // 49 for RGB and others        
	    }
	    
	    // LowResPrecal is lower resolution
	    if ((flags & CMM.cmsFLAGS_LOWRESPRECALC) != 0)
	    {
	        if (nChannels > 4)
	        {
	        	return 6; // 6 for more than 4 channels
	        }
	        
	        if (nChannels == 1)
	        {
	        	return 33; // For monochrome
	        }
	        
	        return 17; // 17 for remaining
	    }
	    
	    // Default values
	    if (nChannels > 4)
	    {
	    	return 7; // 7 for Hifi
	    }
	    
	    if (nChannels == 4)
	    {
	    	return 17; // 17 for CMYK
	    }
	    
	    return 33; // 33 for RGB   
	}
	
	// Chain several profiles into a single LUT. It just checks the parameters and then calls the handler
	// for the first intent in chain. The handler may be user-defined. Is up to the handler to deal with the 
	// rest of intents in chain. A maximum of 255 profiles at time are supported, which is pretty reasonable.
	public static ICC_Pipeline cmsLinkProfiles(int[] intents, ICC_Profile[] profiles, boolean[] bpc, double[] adaptationStates, int flags)
	{
		int nProfiles = profiles.length;
		
		// Make sure a reasonable number of profiles is provided
	    if(nProfiles <= 0 || nProfiles > 255)
	    {
	    	throw new IllegalArgumentException("Couldn't link '" + nProfiles + "' profiles");
	    }
	    
	    for(int i = 0; i < nProfiles; i++)
	    {
	        // Check if black point is really needed or allowed. Note that 
	        // following Adobe's document:
	        // BPC does not apply to devicelink profiles, nor to abs colorimetric, 
	        // and applies always on V4 perceptual and saturation.
	    	
	        if(intents[i] == ICC_Profile.icAbsoluteColorimetric)
	        {
	        	bpc[i] = false;
	        }

	        if(intents[i] == ICC_Profile.icPerceptual || intents[i] == ICC_Profile.icSaturation)
	        {
	        	// Force BPC for V4 profiles in perceptual and saturation
	            if(CMM.cmsGetProfileVersion(profiles[i]) >= 4)
	            {
	            	bpc[i] = true;
	            }
	        }      
	    }
	    
	    // Search for a handler. The first intent in the chain defines the handler. That would
	    // prevent using multiple custom intents in a multiintent chain, but the behaviour of
	    // this case would present some issues if the custom intent tries to do things like
	    // preserve primaries. This solution is not perfect, but works well on most cases.
	    
	    IntentList intent = IntentList.searchIntent(intents[0]);
		if (intent == null)
		{
			throw new IllegalArgumentException("Unsupported intent '" + intents[0] + "'");
		}
		
	    // Call the handler
	    return intent.link(intents, profiles, bpc, adaptationStates, flags);
	}
	
	// This function may be used to set the optional evalueator and a block of private data. If private data is being used, an optional
	// duplicator and free functions should also be specified in order to duplicate the LUT construct. Use NULL to inhibit such functionality.
	private static void _cmsPipelineSetOptimizationParameters(ICC_Pipeline lut, Evaluate16 eval16, Object privateData)
	{
		lut.eval16Fn = eval16;
		lut.data = privateData;
	}
	
	// Black point compensation. Implemented as a linear scaling in XYZ. Black points 
	// should come relative to the white point. Fills an matrix/offset element m
	// which is organized as a 4x4 matrix.
	public static void computeBlackPointCompensation(VEC3 BlackPointIn, VEC3 BlackPointOut, MAT3 m, VEC3 off)
	{
		// Now we need to compute a matrix plus an offset m and of such of
		// [m]*bpin + off = bpout
		// [m]*D50  + off = D50
		//
		// This is a linear scaling in the form ax+b, where
		// a = (bpout - D50) / (bpin - D50)
		// b = - D50* (bpout - bpin) / (bpin - D50)
		
		VEC3 d50_XYZ = CMM.cmsD50_XYZ();
		
		double tx = BlackPointIn.vals[0] - d50_XYZ.vals[0];
		double ty = BlackPointIn.vals[1] - d50_XYZ.vals[1];
		double tz = BlackPointIn.vals[2] - d50_XYZ.vals[2];
		
		double ax = (BlackPointOut.vals[0] - d50_XYZ.vals[0]) / tx;
		double ay = (BlackPointOut.vals[1] - d50_XYZ.vals[1]) / ty;
		double az = (BlackPointOut.vals[2] - d50_XYZ.vals[2]) / tz;
		
		double bx = - d50_XYZ.vals[0] * (BlackPointOut.vals[0] - BlackPointIn.vals[0]) / tx;
		double by = - d50_XYZ.vals[1] * (BlackPointOut.vals[1] - BlackPointIn.vals[1]) / ty;
		double bz = - d50_XYZ.vals[2] * (BlackPointOut.vals[2] - BlackPointIn.vals[2]) / tz;
		
		m.rows[0].set(ax, 0,  0);
		m.rows[1].set(0, ay,  0);
		m.rows[2].set(0,  0,  az);
		off.set(bx, by, bz);
	}
	
	// Approximate a blackbody illuminant based on CHAD information
	private static double CHAD2temp(MAT3 chad)
	{
	    // Convert D50 across CHAD to get the absolute white point
	    VEC3 s = new VEC3(CMM.cmsD50_XYZ());
	    
	    VEC3 d = new VEC3();
	    MAT3.evaluate(d, chad, s);
	    
	    VEC3 dest = new VEC3(d);
	    
	    VEC3 destChromaticity = new VEC3();
	    CMM.cmsXYZ2xyY(destChromaticity, dest);
	    
	    double[] tempK = new double[1];
	    if (!CMM.cmsTempFromWhitePoint(tempK, destChromaticity))
	    {
	    	return -1;
	    }
	    
	    return tempK[0];
	}

	// Compute a CHAD based on a given temperature
	private static void temp2CHAD(MAT3 chad, double temp)
	{
	    VEC3 white = new VEC3();
	    VEC3 chromaticityOfWhite = new VEC3();
	    
	    CMM.cmsWhitePointFromTemp(chromaticityOfWhite, temp);  
	    CMM.cmsxyY2XYZ(white, chromaticityOfWhite);
	    CMM.cmsAdaptationMatrix(chad, null, CMM.cmsD50_XYZ(), white);
	}
	
	// Join scalings to obtain relative input to absolute and then to relative output.
	// Result is stored in a 3x3 matrix
	private static boolean computeAbsoluteIntent(double adaptationState, VEC3 whitePointIn, MAT3 chromaticAdaptationMatrixIn, VEC3 whitePointOut, 
			MAT3 chromaticAdaptationMatrixOut, MAT3 m)
	{
	    // Adaptation state
	    if (adaptationState == 1)
	    {
	    	// Observer is fully adapted. Keep chromatic adaptation. 
	        // That is the standard V4 behaviour
	    	m.rows[0].set(whitePointIn.vals[0] / whitePointOut.vals[0], 0, 0);
	    	m.rows[1].set(0, whitePointIn.vals[1] / whitePointOut.vals[1], 0);
	    	m.rows[2].set(0, 0, whitePointIn.vals[2] / whitePointOut.vals[2]);
	    }
	    else
	    {
	    	// Incomplete adaptation. This is an advanced feature.
	    	MAT3 scale = new MAT3(
	    			new VEC3(whitePointIn.vals[0] / whitePointOut.vals[0], 0, 0), 
	    			new VEC3(0, whitePointIn.vals[1] / whitePointOut.vals[1], 0), 
	    			new VEC3(0, 0, whitePointIn.vals[2] / whitePointOut.vals[2]));
	        
	    	MAT3 m1 = new MAT3(chromaticAdaptationMatrixIn);
	    	MAT3 m2 = new MAT3();
	        if (!MAT3.inverse(m1, m2))
	        {
	        	return false;
	        }
	        MAT3 m3 = new MAT3();
	        MAT3.multiply(m3, m2, scale);
	        
	        // m3 holds CHAD from input white to D50 times abs. col. scaling
	        if (adaptationState == 0)
	        {
	        	// Observer is not adapted, undo the chromatic adaptation
	        	MAT3.multiply(m, m3, chromaticAdaptationMatrixOut);
	        }
	        else
	        {
	            double tempSrc = CHAD2temp(chromaticAdaptationMatrixIn); // K for source white
	            double tempDest = CHAD2temp(chromaticAdaptationMatrixOut); // K for dest white
	            
	            if(tempSrc < 0.0 || tempDest < 0.0)
	            {
	            	return false; // Something went wrong
	            }
	            
	            if(scale.isIdentity() && Math.abs(tempSrc - tempDest) < 0.01)
	            {
	            	m.setIdentity();
	                return true;
	            }
	            
	            double temp = adaptationState * tempSrc + (1 - adaptationState) * tempDest;
	            
	            // Get a CHAD from D50 to whatever output temperature. This replaces output CHAD
	            MAT3 MixedCHAD = new MAT3();
	            temp2CHAD(MixedCHAD, temp);
	            
	            MAT3.multiply(m, m3, MixedCHAD);
	        }
	    }
	    return true;
	}
	
	// Compute the conversion layer
	public static boolean computeConversion(int i, ICC_Profile[] profiles, int intent, boolean bpc, double adaptationState, MAT3 m, VEC3 off)
	{
	    // m  and off are set to identity and this is detected latter on
	    m.setIdentity();
	    off.set(0, 0, 0);

	    // If intent is abs. colorimetric,
	    if (intent == ICC_Profile.icAbsoluteColorimetric)
	    {
	    	VEC3 whitePointIn = new VEC3(), whitePointOut = new VEC3();
	        MAT3 chromaticAdaptationMatrixIn = new MAT3(), chromaticAdaptationMatrixOut = new MAT3();  
	        
	        profiles[i - 1].cmm.cmsReadMediaWhitePoint(whitePointIn);
	        profiles[i - 1].cmm.cmsReadCHAD(chromaticAdaptationMatrixIn);
	        
	        profiles[i].cmm.cmsReadMediaWhitePoint(whitePointOut);
	        profiles[i].cmm.cmsReadCHAD(chromaticAdaptationMatrixOut);
	        
	        if (!computeAbsoluteIntent(adaptationState, whitePointIn,  chromaticAdaptationMatrixIn, whitePointOut, chromaticAdaptationMatrixOut, m))
	        {
	        	return false;
	        }
	    }
	    else
	    {
	    	// Rest of intents may apply BPC.
	    	
	    	if (bpc)
	    	{
	    		VEC3 blackPointIn = new VEC3(), blackPointOut = new VEC3();
	    		
	            CMM.cmsDetectBlackPoint(blackPointIn,  profiles[i - 1], intent, 0);
	            CMM.cmsDetectBlackPoint(blackPointOut, profiles[i], intent, 0);
	            
	            // If black points are equal, then do nothing
	            if (blackPointIn.vals[0] != blackPointOut.vals[0] || blackPointIn.vals[1] != blackPointOut.vals[1] || blackPointIn.vals[2] != blackPointOut.vals[2])
	            {
	            	computeBlackPointCompensation(blackPointIn, blackPointOut, m, off);
	            }
	        }
	    }
	    
	    // Offset should be adjusted because the encoding. We encode XYZ normalized to 0..1.0,
	    // to do that, we divide by MAX_ENCODEABLE_XZY. The conversion stage goes XYZ -> XYZ so
	    // we have first to convert from encoded to XYZ and then convert back to encoded. 
	    // y = Mx + Off
	    // x = x'c
	    // y = M x'c + Off
	    // y = y'c; y' = y / c
	    // y' = (Mx'c + Off) /c = Mx' + (Off / c)
	    
	    for (int k = 0; k < 3; k++)
	    {
	    	off.vals[k] /= ICC_ColorSpace.MAX_XYZ;
	    }
	    
	    return true;
	}
	
	// Add a conversion stage if needed. If a matrix/offset m is given, it applies to XYZ space 
	public static boolean addConversion(ICC_Pipeline result, int inPCS, int outPCS, MAT3 m, VEC3 off)
	{
	    // Handle PCS mismatches. A specialized stage is added to the LUT in such case
	    switch(inPCS)
	    {
	        case ICC_Profile.icSigXYZData: // Input profile operates in XYZ
	            switch (outPCS)
	            {
		            case ICC_Profile.icSigXYZData: // XYZ -> XYZ
		                if (!isEmptyLayer(m, off))
		                {
		                	result.insertStage(new ICC_Stage.Matrix_Stage(3, 3, m.getAsArray(), off.vals), true);
		                }
		                break;
		            case ICC_Profile.icSigLabData: // XYZ -> Lab
		                if (!isEmptyLayer(m, off))
		                {
		                	result.insertStage(new ICC_Stage.Matrix_Stage(3, 3, m.getAsArray(), off.vals), true);
		                }
		                result.insertStage(new ICC_Stage.XYZ2Lab_Stage(), true);
		                break;
		            default:
		                return false; // Colorspace mismatch
                }
                break;
	        case ICC_Profile.icSigLabData: // Input profile operates in Lab
	            switch (outPCS)
	            {
		            case ICC_Profile.icSigXYZData: // Lab -> XYZ
		                result.insertStage(new ICC_Stage.Lab2XYZ_Stage(), true);
		                if(!isEmptyLayer(m, off))
		                {
		                	result.insertStage(new ICC_Stage.Matrix_Stage(3, 3, m.getAsArray(), off.vals), true);
		                }
		                break;
		            case ICC_Profile.icSigLabData: // Lab -> Lab
		                if(!isEmptyLayer(m, off))
		                {
		                	result.insertStage(new ICC_Stage.Lab2XYZ_Stage(), true);
		                	result.insertStage(new ICC_Stage.Matrix_Stage(3, 3, m.getAsArray(), off.vals), true);
		                	result.insertStage(new ICC_Stage.XYZ2Lab_Stage(), true);
		                }
		                break;
		            default:
		                return false; // Mismatch
	            }
	            break;
	            
	            // On colorspaces other than PCS, check for same space
	        default:
	            if (inPCS != outPCS)
	            {
	            	return false;
	            }
	            break;
	    }
	    
	    return true;
	}
	
	// Just to see if m matrix should be applied
	private static boolean isEmptyLayer(MAT3 m, VEC3 off)
	{
	    if (m == null && off == null)
	    {
	    	return true; // NULL is allowed as an empty layer
	    }
	    if (m == null && off != null)
	    {
	    	return false; // This is an internal error
	    }
	    
	    double[] ident = new double[]{1, 0, 0, 0, 1, 0, 0, 0, 1};
	    double[] array = m.getAsArray();
	    double diff = 0;
	    
	    for (int i = 0; i < 3 * 3; i++)
	    {
	    	diff += Math.abs(array[i] - ident[i]);
	    }
	    
	    for (int i = 0; i < 3; i++)
	    {
	    	diff += Math.abs(off.vals[i]);
	    }
	    
	    return (diff < 0.002);
	}
	
	// Is a given space compatible with another?
	private static boolean colorSpaceIsCompatible(int a, int b)
	{
	    // If they are same, they are compatible.
	    if(a == b)
	    {
	    	return true;
	    }
	    
	    // Check for XYZ/Lab. Those spaces are interchangeable as they can be computed one from other.
	    if((a == ICC_Profile.icSigXYZData) && (b == ICC_Profile.icSigLabData))
	    {
	    	return true;
	    }
	    if((a == ICC_Profile.icSigLabData) && (b == ICC_Profile.icSigXYZData))
	    {
	    	return true;
	    }
	    
	    return false;
	}
	
	// Fast floor conversion logic. Thanks to Sree Kotay and Stuart Nixon 
	// note than this only works in the range ..-32767...+32767 because 
	// mantissa is interpreted as 15.16 fixed point.
	// The union is to avoid pointer aliasing overoptimization.
	private static int _cmsQuickFloor(double val)
	{
//#endif
		//The following preprocessor will only get used if USE_LCMS_LIBRARY does not exist
//#ifdef CMS_DONT_USE_FAST_FLOOR
	    return (int)Math.floor(val);
//#endif
//#ifndef USE_LCMS_LIBRARY
	    //Once the proper format is determined (see TODO) then inline double2long function. While at that check bytecode to make sure that it is doing any multiplication.
	    long temp_val = Double.doubleToLongBits(val + (68719476736.0 * 1.5)); //2^36 * 1.5, (52-16=36) uses limited precision to floor
	    
	    return (int)((temp_val & 0xFFFFFFFF) >> 16); //TODO: If this correct or should it be "(int)((temp_val & 0xFFFFFFFF00000000L) >> 48)"?
	}

	// Fast floor restricted to 0..65535.0
	private static short _cmsQuickFloorWord(double d) 
	{ 
	    return (short)(_cmsQuickFloor(d - 32767) + 32767); 
	}

	// Floor to word, taking care of saturation
	private static short _cmsQuickSaturateWord(double d) 
	{
	    d += 0.5;
	    if (d <= 0)
	    {
	    	return 0;
	    }
	    if (d >= 65535)
	    {
	    	return (short)0xFFFF;
	    }
	    
	    return _cmsQuickFloorWord(d);
	}
	
	// Conversion functions. From floating point to 16 bits
	private static void fromFloatTo16(float[] in, short[] out, int n)
	{
	    for (int i = 0; i < n; i++)
	    {
	        out[i] = _cmsQuickSaturateWord(in[i] * ICC_ColorSpace.MAX_SHORT);        
	    }
	}
	
	// From 16 bits to floating point
	private static void from16ToFloat(short[] in, float[] out, int n)
	{
	    for (int i = 0; i < n; i++)
	    {
	        out[i] = (in[i] & 0xFFFF) * ICC_ColorSpace.INV_MAX_SHORT;
	    }
	}
	
	//Intents
	
	private static abstract class IntentList
	{
		private static IntentList defaultIntents;
		
		static
		{
			IntentList intentL;
			
			//Work backwords, only do the first 4 items since this isn't a full implementation of LCMS
			intentL = new DefaultICCIntentList(ICC_Profile.icAbsoluteColorimetric, "Absolute colorimetric", null);
			intentL = new DefaultICCIntentList(ICC_Profile.icSaturation, "Saturation", intentL);
			intentL = new DefaultICCIntentList(ICC_Profile.icRelativeColorimetric, "Relative colorimetric", intentL);
			intentL = new DefaultICCIntentList(ICC_Profile.icPerceptual, "Perceptual", intentL);
			
			IntentList.defaultIntents = intentL;
		}
		
		private static class DefaultICCIntentList extends IntentList
		{
			public DefaultICCIntentList(int intent, String description, IntentList next)
			{
				super(intent, description, next);
			}
			
			//Default handler for ICC-style intents
			public ICC_Pipeline link(int[] intents, ICC_Profile[] profiles, boolean[] bpc, double[] adaptationStates, int flags)
			{
				// For safety
				if(profiles == null || profiles.length == 0)
				{
					return null;
				}
				
				int nProfiles, classSig, intent, colorSpaceIn, colorSpaceOut;
				byte[][] headerCache = new byte[nProfiles = profiles.length][];
				ICC_Pipeline lut, result;
				boolean lIsDeviceLink, lIsInput;
				
				// Allocate an empty LUT for holding the result. 0 as channel count means 'undefined'
				result = new ICC_Pipeline(0, 0);
				
				int currentColorSpace = CMM.getIntFromByteArray(headerCache[0] = profiles[0].getData(ICC_Profile.icSigHead), ICC_Profile.icHdrColorSpace);
				
				for (int i = 0; i < nProfiles; i++)
				{
					ICC_Profile profile = profiles[i];
					if(i != 0)
					{
						headerCache[i] = profile.getData(ICC_Profile.icSigHead);
					}
					
					classSig = CMM.getIntFromByteArray(headerCache[i], ICC_Profile.icHdrDeviceClass);
			        lIsDeviceLink = (classSig == ICC_Profile.icSigLinkClass || classSig == ICC_Profile.icSigAbstractClass);
			        
			        // First profile is used as input unless devicelink or abstract
					if ((i == 0) && !lIsDeviceLink)
					{
						lIsInput = true;
					}
					else
					{
						// Else use profile in the input direction if current space is not PCS
						lIsInput = (currentColorSpace != ICC_Profile.icSigXYZData) && (currentColorSpace != ICC_Profile.icSigLabData);
					}
					
			        intent = intents[i];
			        
			        if (lIsInput || lIsDeviceLink)
			        {
			        	colorSpaceIn = CMM.getIntFromByteArray(headerCache[i], ICC_Profile.icHdrColorSpace);
			            colorSpaceOut = CMM.getIntFromByteArray(headerCache[i], ICC_Profile.icHdrPcs);
			        }
			        else
			        {
			        	colorSpaceIn = CMM.getIntFromByteArray(headerCache[i], ICC_Profile.icHdrPcs);
			            colorSpaceOut = CMM.getIntFromByteArray(headerCache[i], ICC_Profile.icHdrColorSpace);
			        }
			        
			        if (!colorSpaceIsCompatible(colorSpaceIn, currentColorSpace))
			        {
			        	throw new IllegalArgumentException("ColorSpace mismatch");
			        }
			        
			        VEC3 off = new VEC3();
			        MAT3 m = new MAT3();
			        
			        // If devicelink is found, then no custom intent is allowed and we can 
			        // read the LUT to be applied. Settings don't apply here.       
			        if(lIsDeviceLink)
			        {
			        	// Get the involved LUT from the profile
			            lut = new ICC_Pipeline(profile, intent, ICC_Pipeline.READ_DEVICE_LINK);
			            
			            // What about abstract profiles?
			            if(classSig == ICC_Profile.icSigAbstractClass && i > 0)
			            {
			            	if (!computeConversion(i, profiles, intent, bpc[i], adaptationStates[i], m, off))
			            	{
			            		return null;
			            	}
			            }
			            else
			            {
			            	m.setIdentity();
			            	off.set(0, 0, 0);
			            }
			            
			            if(!addConversion(result, currentColorSpace, colorSpaceIn, m, off))
			            {
			            	return null;
			            }
			        }
			        else
			        {
			        	if(lIsInput)
			        	{
			                // Input direction means non-pcs connection, so proceed like devicelinks
			                lut = new ICC_Pipeline(profile, intent, ICC_Pipeline.READ_INPUT);
			            }
			            else
			            {
			            	// Output direction means PCS connection. Intent may apply here
			                lut = new ICC_Pipeline(profile, intent, ICC_Pipeline.READ_OUTPUT);
			                
			                if (!computeConversion(i, profiles, intent, bpc[i], adaptationStates[i], m, off))
			                {
			                	return null;
			                }
			                if (!addConversion(result, currentColorSpace, colorSpaceIn, m, off))
			                {
			                	return null;
			                }
			            }
			        }
			        
			        // Concatenate to the output LUT
			        result.concatenate(lut);
			        
			        // Update current space
			        currentColorSpace = colorSpaceOut;
				}
				
				return result;
			}
		}
		
		private int intent;
		public final String description;
		private IntentList next;
		
		public IntentList(int intent, String description, IntentList next)
		{
			this.intent = intent;
			this.description = description;
			this.next = next;
		}
		
		//Search the list for a suitable intent. Returns null if not found
		public static IntentList searchIntent(int intent)
		{
			for(IntentList pt = defaultIntents; pt != null; pt = pt.next)
			{
				if(pt.intent == intent)
				{
					return pt;
				}
			}
			return null;
		}
		
		// Custom intents. This function should join all profiles specified in the array in
		// a single LUT. Any custom intent in the chain redirects to custom function. If more than
		// one custom intent is found, the one located first is invoked. Usually users should use only one
		// custom intent, so mixing custom intents in same multiprofile transform is not supported.
		
		public abstract ICC_Pipeline link(int[] intents, ICC_Profile[] profiles, boolean[] bpc, double[] adaptationStates, int flags);
	}
	
	static interface Evaluate16
	{
		public void evaluate(short[] in, short[] out);
	}
	
	static interface EvaluateFloat
	{
		public void evaluate(float[] in, float[] out);
	}
	
	private static interface Optimization
	{
		public boolean optimize(ICC_Pipeline lut, int intent, int inputFormat, int outputFormat, int flags);
	}
	
	private static class OptimizationCollection
	{
		public Optimization opt;
		public OptimizationCollection next;
		
		public OptimizationCollection(Optimization opt, OptimizationCollection next)
		{
			this.opt = opt;
			this.next = next;
		}
	}
}

//#endif
