//#preprocessor

//#ifndef USE_LCMS_LIBRARY

/*
 * File: ICC_Stage.java
 * Version: 2.0
 * Initial Creation: Jul 4, 2010 12:02:21 AM
 *
 * SEE "Little CMS License.txt"
 */
package com.sun.pdfview.helper.graphics.color;

/**
 * A stage in ICC conversion pipeline.
 */
public abstract class ICC_Stage
{
	private static final int cmsSigMatrixElemType = 0x6D617466;  // 'matf'
	private static final int cmsSigXYZ2LabElemType = 0x6C327820;  // 'l2x '
	private static final int cmsSigLab2XYZElemType = 0x78326C20;  // 'x2l '
	static final int cmsSigCurveSetElemType = 0x63767374; // 'cvst'
	private static final int cmsSigCLutElemType = 0x636C7574; //'clut'
	
	private ICC_Stage next;
	protected int type, stageImplements, inputChannels, outputChannels;
	protected Object data;
	
	private static final class ICC_InterpParams
	{
		public int flags; // Keep original flags
		public int inputs; // != 1 only in 3D interpolation
		public int outputs; // != 1 only in 3D interpolation
		
		public int[] samples; // Valid on all kinds of tables
		public int[] domain; // Domain = nSamples - 1
		
		public int[] opta; // Optimization for 3D CLUT. This is the number of nodes premultiplied for each
        				   // dimension. For example, in 7 nodes, 7, 7^2 , 7^3, 7^4, etc. On non-regular
        				   // Samplings may vary according of the number of nodes for each dimension.
		
		public Object table; // Points to the actual interpolation table
		
		// This function precalculates as many parameters as possible to speed up the interpolation.
		public ICC_InterpParams(int[] samples, int inputChan, int outputChan, Object table, boolean sixteenBit)
		{
			// Keep original parameters
		    this.flags = sixteenBit ? 1 : 2;
		    this.inputs = inputChan;
		    this.outputs = outputChan;
		    this.table = table;
		    
		    this.samples = new int[inputChan];
		    this.domain = new int[inputChan];
		    this.opta = new int[inputChan];
		    
		    // Fill samples per input direction and domain (which is number of nodes minus one)
		    for (int i = 0; i < inputChan; i++)
		    {
		        this.samples[i] = samples[i];
		        this.domain[i] = samples[i] - 1;     
		    }
		    
		    // Compute factors to apply to each component to index the grid array
		    this.opta[0] = this.outputs;  
		    for (int i = 0; i < inputChan; i++)
		    {
		    	this.opta[i] = this.opta[i - 1] * samples[inputChan - i];
		    }
		    
		    if(!_cmsSetInterpolationRoutine())
		    {
		    	throw new UnsupportedOperationException("Unsupported interpolation (" + inputChan + "->" + outputChan + " channels)");
		    }
		}
		
		// Set the interpolation method
		private boolean _cmsSetInterpolationRoutine()
		{
			//TODO: _cmsSetInterpolationRoutine:cmsintrp.c:#60
			return false;
		}
	}
	
	protected ICC_Stage(int type, int inputChannels, int outputChannels, Object data)
	{
		this.type = type;
		this.stageImplements = type; //By default, no clue on what is implementing
		
		this.inputChannels = inputChannels;
		this.outputChannels = outputChannels;
		
		this.data = data;
	}
	
	//Points to fn that evaluates the stage (always in floating point)
	public abstract void eval(float[] in, float[] out);
	
	public final ICC_Stage next()
	{
		return this.next;
	}
	
	final void setNext(ICC_Stage next)
	{
		this.next = next;
	}
	
	public abstract ICC_Stage clone();
	
	public static final class CLUT_Stage extends ICC_Stage
	{
		public static CLUT_Stage createCLUT_Stage(int gridPoints, int inputChan, int outputChan, Object table, boolean sixteenBit)
		{
			int[] dimensions = new int[inputChan];
			
			// Our resulting LUT would be same gridpoints on all dimensions
			for(int i = 0; i < inputChan; i++)
			{
				dimensions[i] = gridPoints;
			}
			
			return new CLUT_Stage(dimensions, inputChan, outputChan, table, sixteenBit);
		}
		
		public CLUT_Stage(int[] clutPoints, int inputChan, int outputChan, Object table, boolean sixteenBit)
		{
			super(cmsSigCLutElemType, inputChan, outputChan, null);
			
			//Figure out what kind of table to make
			if(table != null)
			{
				if(table instanceof float[] || table instanceof short[])
				{
					if(table instanceof float[] && sixteenBit)
					{
						sixteenBit = false;
					}
					else if(table instanceof short[] && !sixteenBit)
					{
						sixteenBit = false;
					}
				}
				else
				{
					throw new IllegalArgumentException("Unknown table type");
				}
			}
			
			Object[] data = new Object[2];
			int n = outputChan * cubeSize(clutPoints, inputChan);
			if(sixteenBit)
			{
				data[1] = new short[n];
			}
			else
			{
				data[1] = new float[n];
			}
			
			if(table != null)
			{
				System.arraycopy(table, 0, data[1], 0, n);
			}
			
			data[0] = new ICC_InterpParams(clutPoints, inputChan, outputChan, data[1], sixteenBit);
			
			this.data = data;
		}
		
		// Given an hypercube of b dimensions, with Dims[] number of nodes by dimension, calculate the total amount of nodes
		private static int cubeSize(int[] dims, int b)
		{
		    int rv;
		    
		    for (rv = 1; b > 0; b--)
		    {
		    	rv *= dims[b-1];
		    }
		    
		    return rv;
		}
		
		public ICC_Stage clone()
		{
			Object[] data = (Object[])this.data;
			ICC_InterpParams params = (ICC_InterpParams)data[0];
			return new CLUT_Stage(params.samples, this.inputChannels, this.outputChannels, data[1], data[1] instanceof short[]);
		}
		
		public void eval(float[] in, float[] out)
		{
			//TODO
		}
	}
	
	public static final class ToneCurves_Stage extends ICC_Stage
	{
		public static class ToneCurve
		{
			//TODO
			
			public boolean isToneCurveLinear()
			{
				//TODO
				return false;
			}
		}
		
		public ToneCurves_Stage(int nChannels, ToneCurve[] Curves)
		{
			super(cmsSigCurveSetElemType, nChannels, nChannels, null);
			
			//TODO
		}
		
		public ICC_Stage clone()
		{
			return new ToneCurves_Stage(this.outputChannels, (ToneCurve[])this.data);
		}
		
		public ToneCurve[] _cmsStageGetPtrToCurveSet()
		{
			return (ToneCurve[])this.data;
		}
		
		public void eval(float[] in, float[] out)
		{
			//TODO
		}
	}
	
	public static final class Lab2XYZ_Stage extends ICC_Stage
	{
		public Lab2XYZ_Stage()
		{
			super(cmsSigLab2XYZElemType, 3, 3, null);
		}
		
		public ICC_Stage clone()
		{
			return new Lab2XYZ_Stage(); //No parameters so just return a new one
		}
		
		public void eval(float[] in, float[] out)
		{
			VEC3 XYZ = new VEC3();
		    
		    // V4 rules
			VEC3 Lab = new VEC3(in[0] * 100, in[1] * 255 - 128, in[2] * 255 - 128);
		    
		    CMM.cmsLab2XYZ(null, XYZ, Lab);
		    
		    // From XYZ, range 0..19997 to 0..1.0, note that 1.99997 comes from 0xffff
		    // encoded as 1.15 fixed point, so 1 + (32767.0 / 32768.0)
		    
		    out[0] = (float)(XYZ.vals[0] / ICC_ColorSpace.MAX_XYZ);
		    out[1] = (float)(XYZ.vals[1] / ICC_ColorSpace.MAX_XYZ);
		    out[2] = (float)(XYZ.vals[2] / ICC_ColorSpace.MAX_XYZ);
		}
	}
	
	public static final class XYZ2Lab_Stage extends ICC_Stage
	{
		public XYZ2Lab_Stage()
		{
			super(cmsSigXYZ2LabElemType, 3, 3, null);
		}
		
		public ICC_Stage clone()
		{
			return new XYZ2Lab_Stage(); //No parameters so just return a new one
		}
		
		public void eval(float[] in, float[] out)
		{
			VEC3 Lab = new VEC3();
		    
		    // From 0..1.0 to XYZ
		    
		    VEC3 XYZ = new VEC3(in[0] * ICC_ColorSpace.MAX_XYZ, in[1] * ICC_ColorSpace.MAX_XYZ, in[2] * ICC_ColorSpace.MAX_XYZ);
		    
		    CMM.cmsXYZ2Lab(null, Lab, XYZ);
		    
		    // From V4 Lab to 0..1.0
		    
		    out[0] = (float)(Lab.vals[0] / 100); 
		    out[1] = (float)((Lab.vals[1] + 128) / 255); 
		    out[2] = (float)((Lab.vals[2] + 128) / 255);
		}
	}
	
	public static final class Matrix_Stage extends ICC_Stage
	{
		public Matrix_Stage(int rows, int cols, double[] matrix, double[] offset)
		{
			this(rows, cols, matrix, offset, null, true);
		}
		
		private Matrix_Stage(int rows, int cols, double[] matrix, double[] offset, Object data, boolean create)
		{
			super(cmsSigMatrixElemType, cols, rows, data);
			
			if(create)
			{
				int n = rows * cols;
	
			    // Check for overflow
			    if (n < rows || n < cols)
			    {
			    	throw new IllegalArgumentException("rows/cols overflow");
			    }
			    
			    int flag = n;
			    int c = n;
			    
			    if(offset != null && offset.length > 0)
			    {
			    	flag |= 0x80000000;
			    	c += offset.length;
			    }
			    
			    double[] nData = new double[c + 1];
			    nData[0] = flag;
			    
			    for(int i = 0; i < n; i++)
			    {
			    	nData[i + 1] = matrix[i];
			    }
			    
			    if(c > n)
			    {
			    	for (int i = 0; i < cols; i++)
			    	{
			    		nData[n + i + 1] = offset[i];
			    	}
			    }
			    
			    this.data = nData;
			}
		}
		
		public ICC_Stage clone()
		{
			//Copy matrix
			double[] oData = (double[])this.data;
			double[] nData = new double[oData.length];
			
			System.arraycopy(oData, 0, nData, 0, oData.length);
			
			return new Matrix_Stage(this.inputChannels, this.outputChannels, null, null, nData, false);
		}
		
		// Special care should be taken here because precision loss. A temporary cmsFloat64Number buffer is being used
		public void eval(float[] in, float[] out)
		{
			double[] data = (double[])this.data;
			int offsetLen = (data.length - 1) - (((int)data[0]) & ~0x80000000);
			int offset = (data.length - 1) - offsetLen;
			
		    // Input is already in 0..1.0 notation
		    for (int i = 0; i < this.outputChannels; i++)
		    {
		        double tmp = 0;
		        for (int j = 0; j < this.inputChannels; j++)
		        {
		            tmp += in[j] * data[1 + (i * this.inputChannels + j)];
		        }
		        
		        if (offsetLen > 0)
		        {
		        	tmp += data[1 + i + offset];
		        }
		        
		        out[i] = (float)tmp;     
		    }
		    
		    // Output in 0..1.0 domain
		}
	}
}

//#endif
